#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  3 16:03:02 2023

@author: tom.munoz
"""

#==================================================
#
#    Loudspeaker system simulation class
#
#==================================================
# BEM acoustic sim
from electroacPy.acousticSim.bem import bem
from electroacPy.acousticSim.evaluations import evaluations as evs_bem
from electroacPy.acousticSim.evaluations import getPressure
from electroacPy.acousticSim.postProcessing import postProcess as pp

# PS acoustic sim
from electroacPy.acousticSim.pointSource import pointSource, pointSourceBEM

# Lumped element
from electroacPy.speakerSim.electroAcousticDriver import electroAcousticDriver, loadLPM
from electroacPy.speakerSim.enclosureDesign import speakerBox
from electroacPy.speakerSim.filterDesign import xovers

# Vibrometry
from electroacPy.measurements.laserVibrometry import laserVibrometry_UFF as laser_v_uff
from electroacPy.measurements.laserVibrometry import laserVibrometry as laser_v

# general
from electroacPy.global_ import air
from electroacPy.general.freqop import freq_log10
from electroacPy.general.gain import dB
from electroacPy.general.plot import bempp_grid_mesh

# external libraries
import numpy as np
import matplotlib.pyplot as plt


## CLASS
class loudspeakerSystem:
    """
    Automate the use of acousticSim + speakerSim modules specifically for loudspeaker
    design.
    
    Parameters
    ----------
    frequencyRange: numpy array, optional
        Delimitation of the study. Default is freq_log10(20, 2500, 50).
    
    c: float, optional
        Speed of sound in the propagation medium. Default is 343 m/s.
    
    rho: float, optional
        Medium density. Default if 1.22 kg/m^3
        
    Returns
    -------
    None
    """ 
    def __init__(self, frequencyRange=freq_log10(20, 2500, 50),
                 c=air.c, rho=air.rho, **kwargs):
        # global variables
        self.frequency = frequencyRange
        self.c = c
        self.rho = rho

        # simulations
        self.driver         = {}
        self.enclosure      = {}
        self.vibrometry     = {}
        self.crossover      = {}
        self.acoustic_study = {}
        self.evaluation     = {}
        self.results        = {}

        # help
        self.radiator_id    = {}

        if "compiler_output" in kwargs:
            if kwargs["compiler_output"] is True:
                import os
                os.environ['PYOPENCL_COMPILER_OUTPUT'] = '1'
            else:
                pass

    ## ===============================
    # %% LUMPED ELEMENTS / LOUDSPEAKER
    def lem_driver(self, name, U, Le, Re, Cms, Mms, Rms, Bl, Sd, ref2bem=None):
        """
        Add electro-dynamic driver from its Thiele/Small parameters. Can act as a radiator.
        :param name:
        :param U:
        :param Le:
        :param Re:
        :param Cms:
        :param Mms:
        :param Rms:
        :param Bl:
        :param Sd:
        :param ref2bem:
        :return:
        """
        physics = electroAcousticDriver(U, Le, Re, Cms, Mms, Rms, Bl, Sd,
                                        self.frequency, self.c, self.rho)
        physics.ref2bem = ref2bem
        self.driver[name] = physics
        self.radiator_id[name] = 'EAD'
        return None

    def lem_driverImport(self, name, lpm_data, U=1, ref2bem=None, 
                           LeZero=False):
        """
        Add electro-dynamic driver from a LPM file (Thiele/Small parameters from Klippel bench). Can act as a radiator
        :param name:
        :param lpm_data:
        :param U:
        :param ref2bem:
        :return:
        """
        physics = loadLPM(lpm_data, self.frequency, U=U, c=self.c, rho=self.rho, LeZero=LeZero)
        physics.ref2bem = ref2bem
        self.driver[name] = physics
        self.radiator_id[name] = 'EAD'
        return None

    def lem_velocity(self, name, ref2bem, U=1, v=1):
        physics = electroAcousticDriver(U, 1, 1, 1, 1, 1, 1, 1,
                                        self.frequency, self.c, self.rho) # "false" loudspeaker
        physics.ref2bem = ref2bem
        physics.v = np.ones(len(self.frequency), dtype=complex) * v
        physics.Q = np.ones(len(self.frequency), dtype=complex) * v
        self.driver[name] = physics
        self.radiator_id[name] = 'EAD'
        return None

    def lem_enclosure(self, name, Vb, Qab=120, Qal=30, setDriver=None, Nd=1, 
                      wiring="parallel", ref2bem=None, **kwargs):
        
        physics = speakerBox(Vb, frequencyRange=self.frequency, 
                             c=self.c, rho=self.rho,
                             Qab=Qab, Qal=Qal, **kwargs)
        physics.ref2bem = ref2bem
        self.enclosure[name] = physics
        self.radiator_id[name] = 'SPKBOX'

        if setDriver != None:
            physics.getDriverResponse(self.driver[setDriver], 
                                      Nd=Nd, wiring=wiring)
            physics.whichDriver = setDriver
        return None

    def vibrometry_data(self, name, file_path, rotation=[0, 0, 0],
                        ref2bem=None, useAverage=False, inputVoltage=1):
        """
        Add acceleration data to the study. Meant to be used as a radiator - ref2bem strongly recommended.

        Parameters
        ----------
        name : str
            Reference.
        file_path : str
            Path to *.uff vibrometry data.
        rotation : list of float, optional
            Rotate measured data to match mesh direction. The default is [0, 0, 0].
        ref2bem : int, optional
            Reference to mesh physical group. The default is None.
        useAverage : Bool, optional
            Uses the average instead of individual element acceleration. 
            The default is False.
        inputVoltage : float, optional
            Optional scaling. By default, electroacPy uses the transfer function of the 
            acceleration, hence, the input voltage of the laser vibrometer is not taken 
            into account. The default is 1.

        Returns
        -------
        None.

        """
        physics = laser_v_uff(file_path, rotation, self.frequency, useAverage, inputVoltage=inputVoltage)
        physics.ref2bem = ref2bem
        self.vibrometry[name] = physics
        self.radiator_id[name] = 'PLV'
        return None
    
    def vibrometry_data_user(self, name, Hv, X, ref2bem=None, inputVoltage=1):
        """
        Add acceleration data to the study. Meant to be used as a radiator - ref2bem strongly recommended.

        Parameters
        ----------
        name : str
            Reference.
        file_path : str
            Path to *.uff vibrometry data.
        rotation : list of float, optional
            Rotate measured data to match mesh direction. The default is [0, 0, 0].
        ref2bem : int, optional
            Reference to mesh physical group. The default is None.
        useAverage : Bool, optional
            Uses the average instead of individual element acceleration. 
            The default is False.
        inputVoltage : float, optional
            Optional scaling. By default, electroacPy uses the transfer function of the 
            acceleration, hence, the input voltage of the laser vibrometer is not taken 
            into account. The default is 1.

        Returns
        -------
        None.

        """
        physics = laser_v(Hv, X, inputVoltage)
        physics.ref2bem = ref2bem
        self.vibrometry[name] = physics
        self.radiator_id[name] = 'PLV'
        return None

    ## ========================
    # %% FILTERING / CROSSOVERS
    def filter_network(self, name, ref2bem=None, ref2study=None):
        """
        Add a filter network. If linked to a specific study, the correspondind surface should be linked in bem as well.
        :param name:
        :param ref2bem:
        :param ref2study:
        :return:
        """
        self.crossover[name] = xovers(self.frequency)
        self.crossover[name].ref2bem = ref2bem
        self.crossover[name].referenceStudy = ref2study
        return None

    def filter_addLowPass(self, networkRef, filterName, order, fc):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addLowPass(filterName, order, fc)
        else:
            self.crossover[networkRef].addLowPass(filterName, order, fc)
        return None

    def filter_addHighPass(self, networkRef, filterName, order, fc):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addHighPass(filterName, order, fc)
        else:
            self.crossover[networkRef].addHighPass(filterName, order, fc)
        return None

    def filter_addBandPass(self, networkRef, filterName, order, fc1, fc2):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addBandPass(filterName, order, fc1, fc2)
        else:
            self.crossover[networkRef].addBandPass(filterName, order, fc1, fc2)
        return None

    def filter_addLowPassBQ(self, networkRef, filterName, fc, Q, dBGain=0):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addLowPassBQ(filterName, fc, Q, dBGain)
        else:
            self.crossover[networkRef].addLowPassBQ(filterName, fc, Q, dBGain)
        return None

    def filter_addHighPassBQ(self, networkRef, filterName, fc, Q, dBGain=0):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addHighPassBQ(filterName, fc, Q, dBGain)
        else:
            self.crossover[networkRef].addHighPassBQ(filterName, fc, Q, dBGain)
        return None

    def filter_addBandPassBQ(self, networkRef, filterName, fc, Q, dBGain=0):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addBandPassBQ(filterName, fc, Q, dBGain)
        else:
            self.crossover[networkRef].addBandPassBQ(filterName, fc, Q, dBGain)
        return None

    def filter_addPeakEQ(self, networkRef, filterName, fc, Q, dBGain=0):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addPeakEQ(filterName, fc, Q, dBGain)
        else:
            self.crossover[networkRef].addPeakEQ(filterName, fc, Q, dBGain)
        return None

    def filter_addLowShelf(self, networkRef, filterName, fc, Q, dBGain=0):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addLowShelf(filterName, fc, Q, dBGain)
        else:
            self.crossover[networkRef].addLowShelf(filterName, fc, Q, dBGain)
        return None

    def filter_addHighShelf(self, networkRef, filterName, fc, Q, dBGain=0):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addHighShelf(filterName, fc, Q, dBGain)
        else:
            self.crossover[networkRef].addHighShelf(filterName, fc, Q, dBGain)
        return None

    def filter_addTransferFunction(self, networkRef, filterName, tf):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addTransferFunction(filterName, tf)
        else:
            self.crossover[networkRef].addTransferFunction(filterName, tf)
        return None

    def filter_addGain(self, networkRef, filterName, dBGain):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addGain(filterName, dBGain)
        else:
            self.crossover[networkRef].addGain(filterName, dBGain)
        return None

    def filter_addDelay(self, networkRef, filterName, dt, timeConvention='-jwt'):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addDelay(filterName, dt, timeConvention)
        else:
            self.crossover[networkRef].addDelay(filterName, dt, timeConvention)
        return None

    def filter_addPhaseFlip(self, networkRef, filterName):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].addPhaseFlip(filterName)
        else:
            self.crossover[networkRef].addPhaseFlip(filterName)
        return None

    def filter_delete(self, networkRef, filterName):
        if isinstance(networkRef, list):
            for net in networkRef:
                self.crossover[net].deleteFilter(filterName)
        else:
            self.crossover[networkRef].deleteFilter(filterName)
        return None

    ## ==================
    # %% ACOUSTIC STUDIES
    def study_acousticBEM(self, name, meshPath, acoustic_radiator, 
                          domain="exterior", **kwargs):
        """
        Add a boundary element analysis to the loudspeaker study.

        :param name:
        :param meshPath:
        :param acoustic_radiator: obj
            can either be an enclosure or driver object
        :param boundary:
        :param offset:
        :return:
        """
        rad_surf          = []
        surf_v            = []
        vibrometry_points = []
        Nfft              = len(self.frequency)
        if isinstance(acoustic_radiator, str):
            acoustic_radiator = [acoustic_radiator]   # set radiator in list is single str given

        # associate acoustic_radiator to velocity data // vibrometric data
        for i in range(len(acoustic_radiator)):
            cname = acoustic_radiator[i]

            # ========
            # enclosure
            if self.radiator_id[cname] == 'SPKBOX':
                try:
                    nRef = len(self.enclosure[cname].ref2bem)
                    for i in range(nRef):
                        rad_surf.append(self.enclosure[cname].ref2bem[i])
                        surf_v.append(np.ones(Nfft))
                        vibrometry_points.append(False)
                except:
                    rad_surf.append(self.enclosure[cname].ref2bem)
                    surf_v.append(np.ones(Nfft))
                    vibrometry_points.append(False)

            # ========
            # driver
            elif self.radiator_id[cname] == 'EAD':
                try:
                    nRef = len(self.driver[cname].ref2bem)
                    for i in range(nRef):
                        rad_surf.append(self.driver[cname].ref2bem[i])
                        surf_v.append(np.ones(Nfft))
                        vibrometry_points.append(False)
                except:
                    rad_surf.append(self.driver[cname].ref2bem)
                    surf_v.append(np.ones(Nfft))
                    vibrometry_points.append(False)

            # =========================
            # Polytech vibrometric data
            elif self.radiator_id[cname] == 'PLV':
                try:
                    nRef = len(self.vibrometry[cname].ref2bem)
                    for i in range(nRef):
                        rad_surf.append(self.vibrometry[cname].ref2bem[i])
                        surf_v.append(self.vibrometry[cname].v_point)
                        vibrometry_points.append(self.vibrometry[cname].point_cloud)
                except:
                    rad_surf.append(self.vibrometry[cname].ref2bem)
                    surf_v.append(self.vibrometry[cname].v_point)
                    vibrometry_points.append(self.vibrometry[cname].point_cloud)


        # surf_v = np.ones([len(rad_surf), len(self.frequency)])
        physics = bem(meshPath, rad_surf, surf_v, self.frequency, domain,
                      c_0=self.c, rho_0=self.rho, 
                      vibrometry_points=vibrometry_points, **kwargs)
        physics.radiator = acoustic_radiator
        self.acoustic_study[name] = physics
        self.evaluation[name] = evs_bem(physics)
        self.evaluation[name].referenceStudy = name
        return None

    
    def study_acousticPointSource(self, name, xSource, acoustic_radiator, 
                                  meshPath=None, domain="exterior", **kwargs):
        
        if isinstance(acoustic_radiator, str):
            acoustic_radiator = [acoustic_radiator]
        
        rad_point = []
        for cname in acoustic_radiator:
            if self.radiator_id[cname] == 'SPKBOX':
                try:
                    nRef = len(self.enclosure[cname].ref2bem)
                    for i in range(nRef):
                        rad_point.append(self.enclosure[cname].ref2bem[i])
                except:
                    rad_point.append(self.enclosure[cname].ref2bem)

            # ========
            # driver
            elif self.radiator_id[cname] == 'EAD':
                try:
                    nRef = len(self.driver[cname].ref2bem)
                    for i in range(nRef):
                        rad_point.append(self.driver[cname].ref2bem[i])
                except:
                    rad_point.append(self.driver[cname].ref2bem)
            
            
        if meshPath is not None:
            physics = pointSourceBEM(meshPath, xSource, 
                                     np.ones([len(xSource), 
                                              len(self.frequency)]), 
                                     self.frequency, domain=domain, 
                                     c_0=self.c, rho_0=self.rho, 
                                     radiatingElement=rad_point, 
                                     **kwargs)
            physics.radiator = acoustic_radiator
            self.acoustic_study[name] = physics
            self.evaluation[name] = evs_bem(physics)
            self.evaluation[name].referenceStudy = name
        else:
            physics = pointSource(xSource, np.ones([len(xSource), 
                                                    len(self.frequency)]), 
                                  self.frequency, radiatingElement=rad_point,
                                  c=self.c, rho=self.rho, **kwargs)
            physics.radiator = acoustic_radiator
            self.acoustic_study[name] = physics
            self.evaluation[name] = evs_bem(physics)
            self.evaluation[name].referenceStudy = name
        return None


    ## =======================
    # %% ACOUSTIC obs
    def evaluation_polarRadiation(self,
                           reference_study: str or list,
                           evaluation_name: str,
                           min_angle: float,
                           max_angle: float,
                           step: float,
                           on_axis: str,
                           direction: str,
                           radius: float = 5,
                           offset: list = [0, 0, 0]):
        """
        Add a circular microphone array to given study.

        :param reference_study:
        :param evaluation_name:
        :return:
        """
        if isinstance(reference_study, list):
            for i in range(len(reference_study)):
                self.evaluation[reference_study[i]].polarRadiation(evaluation_name,
                                                                    min_angle,
                                                                    max_angle,
                                                                    step,
                                                                    on_axis,
                                                                    direction,
                                                                    radius,
                                                                    offset)
        else:
            self.evaluation[reference_study].polarRadiation(evaluation_name,
                                                                min_angle,
                                                                max_angle,
                                                                step,
                                                                on_axis,
                                                                direction,
                                                                radius,
                                                                offset)
        return None

    def evaluation_pressureField(self,
                          reference_study: str,
                          evaluation_name: str,
                          L1: float,
                          L2: float,
                          step: float,
                          plane: str,
                          offset: list = [0, 0, 0],
                          addToPlotter: bool = True):
        """
        Add planar microphone array for "slice" pressure plot in given study.

        :param reference_study:
        :param evaluation_name:
        :param L1:
        :param L2:
        :param step:
        :param plane:
        :param offset:
        :param addToPlotter:
        :return:
        """

        if isinstance(reference_study, list):
            for i in range(len(reference_study)):
                self.evaluation[reference_study[i]].pressureField(evaluation_name,
                                                                   L1,
                                                                   L2,
                                                                   step,
                                                                   plane,
                                                                   offset)
        else:
            self.evaluation[reference_study].pressureField(evaluation_name,
                                                               L1,
                                                               L2,
                                                               step,
                                                               plane,
                                                               offset)
        return None

    def evaluation_fieldPoint(self,
                             reference_study: str,
                             evaluation_name: str,
                             microphonePositions: list,
                             **kwargs):
        """
        Add FRF evaluation points at defined position on given study.

        :param reference_study:
        :param evaluation_name:
        :param microphonePositions:
        :param labels:
        :return:
        """

        if isinstance(reference_study, list):
            for i in range(len(reference_study)):
                self.evaluation[reference_study[i]].fieldPoint(evaluation_name,
                                                                microphonePositions,
                                                                **kwargs)
        else:
            self.evaluation[reference_study].fieldPoint(evaluation_name,
                                                         microphonePositions,
                                                         **kwargs)
        return None

    def evaluation_boundingBox(self,
                                reference_study: str,
                                evaluation_name: str,
                                Lx: float,
                                Ly: float,
                                Lz: float,
                                step: float = 343/500/5,
                                offset: list = [0, 0, 0]):

        if isinstance(reference_study, list):
            for i in range(len(reference_study)):
                self.evaluation[reference_study[i]].boundingBox(evaluation_name,
                                                                 Lx, Ly, Lz, step, offset)
        else:
            self.evaluation[reference_study].boundingBox(evaluation_name,
                                                          Lx, Ly, Lz, step, offset)
        return None

    def evaluation_sphericalRadiation(self,
                                       reference_study: str,
                                       evaluation_name: str,
                                       nMic: float,
                                       radius: float = 1.8,
                                       offset: list = [0, 0, 0]):

        if isinstance(reference_study, list):
            for i in range(len(reference_study)):
                self.evaluation[reference_study[i]].sphericalRadiation(evaluation_name,
                                                                 nMic, radius, offset)
        else:
            self.evaluation[reference_study].sphericalRadiation(evaluation_name,
                                                          nMic, radius, offset)
        return None

    
    def evaluation_plottingGrid(self, 
                                reference_study: str, 
                                evaluation_name: str,
                                path_to_grid: str):
        
        if isinstance(reference_study, list):
            for i in range(len(reference_study)):
                self.evaluation[reference_study[i]].plottingGrid(evaluation_name,
                                                                 path_to_grid)
        else:
            self.evaluation[reference_study].plottingGrid(evaluation_name,
                                                          path_to_grid)

    ## ===================
    # %% run / plot / info
    def run(self):
        """
        Run all defined studies

        """

        for study in self.acoustic_study:
            if self.acoustic_study[study].isComputed is False:
                if hasattr(self.acoustic_study[study], "xSource") and self.acoustic_study[study].meshPath is None:
                    pass # check if it is a point source study without BEM boundaries. No "solve" if True
                else:
                    self.acoustic_study[study].solve()
            else:
                None
            
        for obs in self.evaluation:
            if bool(self.evaluation[obs].setup) is True:
                self.evaluation[obs].solve()
            else:
                print("No evaluation to compute for study {}.".format(obs))
        return None

    ## PLOT
    def plot_results(self, study=[], 
                     evaluation=[], radiatingElement=[], bypass_xover=False,
                     transformation="SPL", export_grid=False, pf2grid=False):
        
        # update solutions
        if isinstance(radiatingElement, int) is True: # avoid possible error if only one rad surf is selected
            radiatingElement = [radiatingElement]
                
        if bool(study) is False: # if not specific study given, plot all
            for s in self.acoustic_study:
                _ = updateResults(self, s, bypass_xover)
                _ = self.evaluation[s].plot(evaluation, radiatingElement,  
                                             processing=self.results[s],
                                             transformation=transformation,
                                             export_grid=export_grid, 
                                             pf2grid=pf2grid)
        else: # plot specific study
            _ = updateResults(self, study, bypass_xover)
            _ = self.evaluation[study].plot(evaluation, radiatingElement,
                                                 processing=self.results[study],
                                                 transformation=transformation,
                                                 export_grid=export_grid,
                                                 pf2grid=pf2grid)
        return None

    def plot_system(self, study, backend="pyvista"):
        """
        Plot study's mesh and related evaluations. By default it uses PyVista as the 
        plotting backend, but can use Gmsh if "gmsh" is passed.

        Parameters
        ----------
        study : str
            Study to display.
        backend : str, optional
            Which backend should be used. The default is "pyvista".

        Returns
        -------
        A plot with system mesh and evaluation points.

        """
        self.evaluation[study].plot_system(backend=backend)
        return None

    def plot_xovers(self, networks, split=True, amplitude=64):
        h_p = []
        names = []
        h = np.zeros(len(self.frequency), dtype=complex)
        if isinstance(networks, str) is True:
            self.crossover[networks].plot(divide_stages=split, amplitude=amplitude)
        else:
            for i in range(len(networks)):
                self.crossover[networks[i]].plot(divide_stages=split, amplitude=amplitude)
                h += self.crossover[networks[i]].h
                h_p.append(self.crossover[networks[i]].h)
                names.append(networks[i])
            plot_network(self, h, h_p, names, split, amplitude)
        return
    
    def plot_pressureMesh(self, study, radiatingElement=[],
                          bypass_xover=False, transformation="SPL", 
                          export_grid=False, backend="gmsh"):
        
        # update solutions
        if isinstance(radiatingElement, int) is True: # avoid possible error if only one rad surf is selected
            radiatingElement = [radiatingElement]
            
        if bool(radiatingElement) is False:
            # plot all elements
            element2plot = self.acoustic_study[study].radiatingElement
        else:
            element2plot = radiatingElement
                
        _ = updateResults(self, study, bypass_xover)        
        if backend == "gmsh":
            elementCoeff = np.ones([len(self.frequency), len(element2plot)], 
                                   dtype=complex)
            pp = self.results[study]
            for name in pp.TF:
                for idx, element in enumerate(element2plot):
                    if element in pp.TF[name]["radiatingElement"]:
                        elementCoeff[:, idx] *= pp.TF[name]["H"]
            
            bempp_grid_mesh(self.acoustic_study[study], elementCoeff, 
                            element2plot, transformation, export_grid)          
        return None


    ## Get Values
    # INFO
    def info(self):
        """
        Returns informations about the current simulation.

        Returns
        -------
        None
        """
        out_message = ()
        maxInfo = []
        if bool(self.driver) is True:
            ead_INFO = list(self.driver)
            out_message += (ead_INFO, )
            maxInfo.append(len(str(ead_INFO)))
        else:
            out_message += ('No driver defined', )
        if bool(self.enclosure) is True:
            box_INFO = list(self.enclosure)
            out_message += (box_INFO,)
            maxInfo.append(len(str(box_INFO)))
        else:
            out_message += ('No enclosure defined', )
        if bool(self.vibrometry) is True:
            laser_INFO = list(self.vibrometry)
            out_message += (laser_INFO, )
            maxInfo.append(len(str(laser_INFO)))
        else:
            out_message += ('No vibration data imported', )
        if bool(self.crossover) is True:
            xover_INFO = list(self.crossover)
            out_message += (xover_INFO,)
            maxInfo.append(len(str(xover_INFO)))
        else:
            out_message += ('No crossover defined',)
        if bool(self.acoustic_study) is True:
            study_INFO = list(self.acoustic_study)
            out_message += (study_INFO,)
            maxInfo.append(len(str(study_INFO)))
        else:
            out_message += ('No study defined',)
        
        maxInfo = np.max(maxInfo)
        maxDesc = len("# Electro-acoustic Drivers")
        maxTot = np.max([maxInfo,  maxDesc])

        print(maxTot * '#')
        print("# Electro-acoustic Drivers")
        print("# " + str(out_message[0]))

        print(maxTot * "-")
        print("# Enclosures")
        print("# " + str(out_message[1]))
        
        print(maxTot * '-')
        print("# Vibrometric data")
        print("# " + str(out_message[2]))

        print(maxTot * "-")
        print("# Crossovers")
        print("# " + str(out_message[3]))

        print(maxTot * "-")
        print("# Acoustical Studies")
        print("# " + str(out_message[4]))
        print(maxTot * "#")
        return None

    def get_pMic(self, studyName, evaluationName, 
                 radiatingElement=[], bypass_xover=False):
        """
        Return pressure at microphone points for defined studies, evaluation and radiating surfaces.

        :param studyName:
        :param evaluationName:
        :param radiatingSurface:
        :param get_freq_array:
        :return:
        """
        if bool(radiatingElement) is False:
            radiatingElement = self.acoustic_study[studyName].radiatingElement 
        
        if isinstance(radiatingElement, int) is True: # avoid possible error if only one rad surf is selected
            radiatingElement = [radiatingElement]

        _  = updateResults(self, studyName, bypass_xover)
        pmic = self.evaluation[studyName].setup[evaluationName].pMic
        
        if self.results[studyName] is not None:
            elementCoeff = np.ones([len(self.frequency), len(radiatingElement)], 
                                   dtype=complex)
            pp = self.results[studyName]
            for name in pp.TF:
                if bypass_xover is False:
                    for idx, element in enumerate(radiatingElement):
                        if element in pp.TF[name]["radiatingElement"]:
                            elementCoeff[:, idx] *= pp.TF[name]["H"]
                elif bypass_xover is True and name[:12] != "filter_stage":
                    for idx, element in enumerate(radiatingElement):
                        if element in pp.TF[name]["radiatingElement"]:
                            elementCoeff[:, idx] *= pp.TF[name]["H"]
        else:
            elementCoeff = np.ones([len(self.frequency), len(radiatingElement)], 
                                   dtype=complex)
        
        out = getPressure(pmic, self.acoustic_study[studyName].radiatingElement,
                                radiatingElement, elementCoeff)
        return out
    
    
    def export_directivity(self, folder_path, file_name,
                           study, evaluation, radiatingElement=[], 
                           bypass_xover=False, frd=False):
        """
        Export directivity results.

        Parameters
        ----------
        folder_name : str
            where data are saved.
        file_name: str
            file prefix.
        study : str
            name of study to export.
        evaluation : str
            evaluation to export.
        radiatingElement : int or list of int, optional
            extract specific element.
        bypass_xover : bool, optional 
            if True, will export filtered response.
        frd : bool, optional
            if True, uses *.frd file extension instead of *.txt

        Returns
        -------
        None.

        """
        from electroacPy.general.acoustics import export_directivity
        
        pmic = self.get_pMic(study, evaluation, radiatingElement, bypass_xover)
        theta = self.evaluation[study].setup[evaluation].theta
        frequency = self.frequency
        export_directivity(folder_path, file_name, frequency, theta, pmic, frd)
        
        
    def export_impedance(self, folder_path, file_name, objName, zma=False):
        """
        Export impedance into .txt file

        Parameters
        ----------
        folder_name : str
            where data are saved.
        file_name: str
            file prefix.
        objName : str
            enclosure or driver object to export.
        zma : bool, optional
            if True, uses the *.zma file extension.
            
        Returns
        -------
        None.

        """
        
        if zma is False:
            if objName in self.enclosure:
                self.enclosure[objName].exportZe(folder_path, file_name + ".txt")
            elif objName in self.driver:
                self.driver[objName].exportZe(folder_path, file_name + ".txt")
        elif zma is True:
            if objName in self.enclosure:
                self.enclosure[objName].exportZe(folder_path, file_name + ".zma")
            elif objName in self.driver:
                self.driver[objName].exportZe(folder_path, file_name + ".zma")
        
    
## =======================
# %% post-processing tools
def updateResults(loudspeakerSystem, study_to_plot, bypass_xover):
    ls = loudspeakerSystem
    study = study_to_plot
    ls.results[study] = pp()
    radiatorName      = ls.acoustic_study[study].radiator
    for name in radiatorName:
        if ls.radiator_id[name] == 'SPKBOX':
            _ = apply_Velocity_From_SPKBOX(ls, study, name)
        elif ls.radiator_id[name] == 'EAD':
            _ = apply_Velocity_From_EAD(ls, study, name)
        elif ls.radiator_id[name] == 'PLV':
            _ = apply_Velocity_From_PLV(ls, study, name)

    if bypass_xover is False:
        for xover in ls.crossover:
            if study in ls.crossover[xover].referenceStudy:  # check if some of the crossovers are to be applied in the study
                h = ls.crossover[xover].h
                ref2bem = ls.crossover[xover].ref2bem
                if isinstance(ref2bem, int):
                    ls.results[study].addTransferFunction("filter_stage_"+xover, 
                                                          h, [ref2bem])
                elif isinstance(ref2bem, list):
                    ls.results[study].addTransferFunction("filter_stage_"+xover, 
                                                          h, ref2bem)
                else:
                    print("you shouldn't use a numpy array, at least for now :)")
    return None

def apply_Velocity_From_SPKBOX(loudspeakerSystem, study, radiatorName):
    """
    Apply velocity from speakerBox object on BEM simulation results

    :param loudspeakerSystem:
    :param radiatorName:
    :param portOnly
    :return:
    """
    ls = loudspeakerSystem
    if "xSource" not in ls.acoustic_study[study].__dict__: # check if point_source
        v = ls.enclosure[radiatorName].v      # driver velocity
        vp = ls.enclosure[radiatorName].vp    # port's velocity
        vp2 = ls.enclosure[radiatorName].vp2  # port 2 velocity (in case of bp2 enclosure config)
        vpr = ls.enclosure[radiatorName].vpr  # passive radiator velocity (in case of pr enclosure config)
        vpr2 = ls.enclosure[radiatorName].vpr2
    else:
        v = ls.enclosure[radiatorName].Q     # driver velocity
        vp = ls.enclosure[radiatorName].Qp    # port's velocity
        vp2 = ls.enclosure[radiatorName].Qp2  # port 2 velocity (in case of bp2 enclosure config)
        vpr = ls.enclosure[radiatorName].Qpr  # passive radiator velocity (in case of pr enclosure config)
        vpr2 = ls.enclosure[radiatorName].Qpr2
    ref2bem = ls.enclosure[radiatorName].ref2bem
        
    if isinstance(ref2bem, int) is True:
        if ls.enclosure[radiatorName].config == "sealed":  # check if 1 radiator because sealed or bp config
            ls.results[study].addTransferFunction("driver_"+radiatorName,
                                                  v, [ref2bem])
        elif ls.enclosure[radiatorName].config == "bandpass":
            ls.results[study].addTransferFunction("port_"+radiatorName, 
                                                  vp, [ref2bem])
        elif ls.enclosure[radiatorName].config == "bandpass_pr":
            ls.results[study].addTransferFunction("passive-radiator_"+"radiatorName",
                                                  vpr, [ref2bem])
    elif isinstance(ref2bem, list) is True:
        if ls.enclosure[radiatorName].config == "sealed":
            ls.results[study].addTransferFunction("driver_"+radiatorName, v, ref2bem)
        elif ls.enclosure[radiatorName].config == "vented":
            ls.results[study].addTransferFunction("driver_"+radiatorName, v, ref2bem[:-1])
            ls.results[study].addTransferFunction("port_"+radiatorName, vp, [ref2bem[-1]])
        
        elif ls.enclosure[radiatorName].config == "passiveRadiator":
            ls.results[study].addTransferFunction("driver_"+radiatorName, v, ref2bem[:-1])
            ls.results[study].addTransferFunction("passive-radiator_"+radiatorName, vpr, [ref2bem[-1]])
        
        elif ls.enclosure[radiatorName].config == "bandpass":
            ls.results[study].addTransferFunction("port_"+radiatorName, vp, ref2bem)
        
        elif ls.enclosure[radiatorName].config == "bandpass_2":
            ls.results[study].addTransferFunction("portf_"+radiatorName, vp, [ref2bem[0]])
            ls.results[study].addTransferFunction("portb_"+radiatorName, vp2, [ref2bem[1]])

        elif ls.enclosure[radiatorName].config == "bandpass_pr":
            ls.results[study].addTransferFunction("passive-radiator_"+radiatorName, vpr, ref2bem)
        
        elif ls.enclosure[radiatorName].config == "bandpass_pr_2":
            ls.results[study].addTransferFunction("prf_"+radiatorName, vpr, [ref2bem[0]])
            ls.results[study].addTransferFunction("prb_"+radiatorName, vpr2, [ref2bem[1]])        
    return None

def apply_Velocity_From_EAD(loudspeakerSystem, study, radiatorName):
    """
    Apply velocity from EAD object on BEM simulation results

    :param loudspeakerSystem:
    :param radiatorName:
    :return:
    """
    ls = loudspeakerSystem
    
    if "xSource" not in ls.acoustic_study[study].__dict__:
        v = ls.driver[radiatorName].v
    else:
        v = ls.driver[radiatorName].Q
    ref2bem = ls.driver[radiatorName].ref2bem
    if isinstance(ref2bem, int) is True:
        ls.results[study].addTransferFunction("driver_"+radiatorName, v, [ref2bem])
    elif isinstance(ref2bem, list) is True:
        ls.results[study].addTransferFunction("driver_"+radiatorName, v, ref2bem)
    return None

def apply_Velocity_From_PLV(loudspeakerSystem, study, radiatorName):
    """
    Apply decoy velocity from PLV object on BEM simulation results

    :param loudspeakerSystem:
    :param radiatorName:
    :return:
    """
    ls = loudspeakerSystem
    v = ls.vibrometry[radiatorName].v
    ref2bem = ls.vibrometry[radiatorName].ref2bem
    if isinstance(ref2bem, int) is True:
        ls.results[study].addTransferFunction("unit_v_"+radiatorName, v, [ref2bem])
    elif isinstance(ref2bem, list) is True:
        ls.results[study].addTransferFunction("unit_v_"+radiatorName, v, ref2bem)
    return None

## ===============
# %% plot function
def plot_network(system, h, h_p, names, split, amplitude):
    if split is False:
        fig, ax = plt.subplots(2, 1)
        ax[0].semilogx(system.frequency, dB(h), 'k')
        ax[1].semilogx(system.frequency, np.angle(h), 'k')
        ax[0].set(xlabel='Frequency [Hz]', ylabel='Gain [dB]',
               title='Total network response',
               ylim=[-amplitude/2, +amplitude/2])
        ax[1].set(xlabel='Frequency [Hz]', ylabel='Phase [rad]')
        for i in range(2):
            ax[i].grid(linestyle='dotted', which='both')
            ax[i].legend(loc='best')
        plt.tight_layout()
    else:
        fig, ax = plt.subplots(2, 1)
        for i in range(len(h_p)):
            ax[0].semilogx(system.frequency, dB(h_p[i]), label=names[i])
            ax[1].semilogx(system.frequency, np.angle(h_p[i]), label=names[i])
        ax[0].semilogx(system.frequency, dB(h), 'k', label='summed response')
        ax[1].semilogx(system.frequency, np.angle(h), 'k', label='summed response')
        ax[0].set(xlabel='Frequency [Hz]', ylabel='Gain [dB]',
               title='Total network response',
               ylim=[-amplitude/2, +amplitude/2])
        ax[1].set(xlabel='Frequency [Hz]', ylabel='Phase [rad]')
        for j in range(2):
            ax[j].grid(linestyle='dotted', which='both')
            ax[j].legend(loc='best')
        plt.tight_layout()
    return plt.show()



"""Sparse assembly."""

import numpy as _np

from bempp_cl.api.assembly import assembler as _assembler
from bempp_cl.helpers import timeit as _timeit


class SparseAssembler(_assembler.AssemblerBase):
    """Implementation of a sparse assembler."""

    # pylint: disable=useless-super-delegation
    def __init__(self, domain, dual_to_range, parameters=None):
        """Create a dense assembler instance."""
        super().__init__(domain, dual_to_range, parameters)

    def assemble(self, operator_descriptor, device_interface, precision, *args, **kwargs):
        """Sparse assembly of operators."""
        from bempp_cl.api.utils.helpers import promote_to_double_precision
        from bempp_cl.api.space.space import return_compatible_representation
        from .numba_kernels import select_numba_kernels
        from scipy.sparse import coo_matrix
        from bempp_cl.api.assembly.discrete_boundary_operator import (
            SparseDiscreteBoundaryOperator,
        )

        domain, dual_to_range = return_compatible_representation(self.domain, self.dual_to_range)
        row_grid_dofs = dual_to_range.grid_dof_count
        col_grid_dofs = domain.grid_dof_count

        if domain.grid != dual_to_range.grid:
            raise ValueError("For sparse operators the domain and dual_to_range grids must be identical.")

        trial_local2global = domain.local2global.ravel()
        test_local2global = dual_to_range.local2global.ravel()
        trial_multipliers = domain.local_multipliers.ravel()
        test_multipliers = dual_to_range.local_multipliers.ravel()

        numba_assembly_function, numba_kernel_function = select_numba_kernels(operator_descriptor, mode="sparse")

        rows, cols, values = assemble_sparse(
            domain.localised_space,
            dual_to_range.localised_space,
            self.parameters,
            operator_descriptor,
            numba_assembly_function,
            numba_kernel_function,
        )
        global_rows = test_local2global[rows]
        global_cols = trial_local2global[cols]
        global_values = values * trial_multipliers[cols] * test_multipliers[rows]

        if self.parameters.assembly.always_promote_to_double:
            values = promote_to_double_precision(values)

        mat = coo_matrix(
            (global_values, (global_rows, global_cols)),
            shape=(row_grid_dofs, col_grid_dofs),
        ).tocsr()

        if domain.requires_dof_transformation:
            mat = mat @ domain.dof_transformation

        if dual_to_range.requires_dof_transformation:
            mat = dual_to_range.dof_transformation.T @ mat

        return SparseDiscreteBoundaryOperator(mat)


@_timeit
def assemble_sparse(
    domain,
    dual_to_range,
    parameters,
    operator_descriptor,
    numba_assembly_function,
    numba_kernel_function,
):
    """Actually assemble the operator."""
    import bempp_cl.api
    from bempp_cl.api.integration.triangle_gauss import rule as regular_rule
    from bempp_cl.api.utils.helpers import get_type

    order = parameters.quadrature.regular
    quad_points, quad_weights = regular_rule(order)

    support = domain.support * dual_to_range.support

    elements = _np.flatnonzero(support)
    number_of_elements = len(elements)

    nshape_test = dual_to_range.number_of_shape_functions
    nshape_trial = domain.number_of_shape_functions

    # Always assemble in double precision for sparse ops
    # precision = operator_descriptor.precision

    precision = "double"

    if operator_descriptor.is_complex:
        result_type = get_type(precision).complex
    else:
        result_type = get_type(precision).real

    result = _np.zeros(nshape_test * nshape_trial * number_of_elements, dtype=result_type)

    if operator_descriptor.identifier == "laplace_beltrami":
        dual_to_range_shapeset_evaluation = dual_to_range.shapeset.gradient
        domain_shapeset_evaluation = domain.shapeset.gradient
        dual_to_range_numba_evaluation = dual_to_range.numba_surface_gradient
        domain_numba_evaluation = domain.numba_surface_gradient
    elif operator_descriptor.identifier == "_vector_grad_product":
        dual_to_range_shapeset_evaluation = dual_to_range.shapeset.evaluate
        domain_shapeset_evaluation = domain.shapeset.gradient
        dual_to_range_numba_evaluation = dual_to_range.numba_evaluate
        domain_numba_evaluation = domain.numba_surface_gradient
    elif operator_descriptor.identifier == "_curl_curl_product":
        dual_to_range_shapeset_evaluation = dual_to_range.shapeset.gradient
        domain_shapeset_evaluation = domain.shapeset.gradient
        dual_to_range_numba_evaluation = dual_to_range.numba_surface_curl
        domain_numba_evaluation = domain.numba_surface_curl
    else:
        dual_to_range_shapeset_evaluation = dual_to_range.shapeset.evaluate
        domain_shapeset_evaluation = domain.shapeset.evaluate
        dual_to_range_numba_evaluation = dual_to_range.numba_evaluate
        domain_numba_evaluation = domain.numba_evaluate

    with bempp_cl.api.Timer() as t:  # noqa: F841
        numba_assembly_function(
            domain.grid.data(precision),
            nshape_test,
            nshape_trial,
            elements,
            quad_points,
            quad_weights,
            dual_to_range.normal_multipliers,
            domain.normal_multipliers,
            dual_to_range.local_multipliers,
            domain.local_multipliers,
            dual_to_range_shapeset_evaluation,
            domain_shapeset_evaluation,
            dual_to_range_numba_evaluation,
            domain_numba_evaluation,
            numba_kernel_function,
            result,
        )

    irange = _np.arange(nshape_test)
    jrange = _np.arange(nshape_trial)

    i_ind = _np.tile(_np.repeat(irange, nshape_trial), len(elements)) + _np.repeat(
        elements * nshape_test,
        nshape_test * nshape_trial,
    )

    j_ind = _np.tile(_np.tile(jrange, nshape_test), len(elements)) + _np.repeat(
        elements * nshape_trial,
        nshape_test * nshape_trial,
    )

    return i_ind, j_ind, result



"""
Basic helpers that should be available independent of the api.

This is for module level helpers that should be available without importing
bempp_cl.api first.

"""

import time as _time
import functools as _functools
import collections as _collections


def timeit(fun):
    """Time a method in Bempp."""

    @_functools.wraps(fun)
    def timed_fun(*args, **kwargs):
        """Time an operation."""
        from bempp_cl.api import log

        start_time = _time.time()
        res = fun(*args, **kwargs)
        end_time = _time.time()
        log(fun.__qualname__ + " : {0:.3e}s".format(end_time - start_time), "timing")
        return res

    return timed_fun


IndexList = _collections.namedtuple("IndexList", ["indices", "indexptr"])


def jit_logger(name):
    """Emit a log message whenever Numba jits something."""
    import bempp_cl.api

    def closure(func):
        """Closure that has the name variable."""

        def inner(*args, **kwargs):
            origsigs = set(func.signatures)
            result = func(*args, **kwargs)
            newsigs = set(func.signatures)
            if newsigs != origsigs:
                new = (newsigs ^ origsigs).pop()
                bempp_cl.api.log(f"Compiled {name} for signature {new}", level="timing")
            return result

        return inner

    return closure



"""Definitions of finite element spaces.

Documentation of the functionality of this module and examples
can be found at https://bempp.com/handbook/api/function_spaces.html
"""

from .space import function_space



"""Actual implementation of Numba assemblers."""

import numpy as _np


def singular_assembler(
    device_interface,
    operator_descriptor,
    grid,
    domain,
    dual_to_range,
    test_points,
    trial_points,
    quad_weights,
    test_elements,
    trial_elements,
    test_offsets,
    trial_offsets,
    weights_offsets,
    number_of_quad_points,
    kernel_options,
    result,
):
    """Numba assembler for the singular part of integral operators."""
    from bempp_cl.api.utils.helpers import get_type
    from bempp_cl.core.numba_kernels import select_numba_kernels

    numba_assembly_function, numba_kernel_function = select_numba_kernels(operator_descriptor, mode="singular")

    # Perform Numba assembly always in double precision
    # precision = operator_descriptor.precision
    precision = "double"
    dtype = get_type(precision).real

    numba_assembly_function(
        grid.data(precision),
        test_points,
        trial_points,
        quad_weights,
        test_elements,
        trial_elements,
        test_offsets,
        trial_offsets,
        weights_offsets,
        number_of_quad_points,
        dual_to_range.normal_multipliers,
        domain.normal_multipliers,
        dual_to_range.number_of_shape_functions,
        domain.number_of_shape_functions,
        dual_to_range.shapeset.evaluate,
        domain.shapeset.evaluate,
        numba_kernel_function,
        _np.array(kernel_options, dtype=dtype),
        result,
    )


def dense_assembler(device_interface, operator_descriptor, domain, dual_to_range, parameters, result):
    """Numba based dense assembler."""
    from bempp_cl.core.numba_kernels import select_numba_kernels
    from bempp_cl.api.utils.helpers import get_type
    from bempp_cl.api.integration.triangle_gauss import rule

    (
        numba_assembly_function_regular,
        numba_kernel_function_regular,
    ) = select_numba_kernels(operator_descriptor, mode="regular")

    order = parameters.quadrature.regular
    quad_points, quad_weights = rule(order)

    # Perform Numba assembly always in double precision
    # precision = operator_descriptor.precision
    precision = "double"

    data_type = get_type(precision).real

    test_indices, test_color_indexptr = dual_to_range.get_elements_by_color()
    trial_indices, trial_color_indexptr = domain.get_elements_by_color()
    number_of_test_colors = len(test_color_indexptr) - 1
    # number_of_trial_colors = len(trial_color_indexptr) - 1

    # rows = dual_to_range.global_dof_count
    # cols = domain.global_dof_count

    nshape_test = dual_to_range.number_of_shape_functions
    nshape_trial = domain.number_of_shape_functions
    grids_identical = domain.grid == dual_to_range.grid

    for test_color_index in range(number_of_test_colors):
        numba_assembly_function_regular(
            dual_to_range.grid.data(precision),
            domain.grid.data(precision),
            nshape_test,
            nshape_trial,
            test_indices[test_color_indexptr[test_color_index] : test_color_indexptr[1 + test_color_index]],
            trial_indices,
            dual_to_range.local_multipliers.astype(data_type),
            domain.local_multipliers.astype(data_type),
            dual_to_range.local2global,
            domain.local2global,
            dual_to_range.normal_multipliers,
            domain.normal_multipliers,
            quad_points.astype(data_type),
            quad_weights.astype(data_type),
            numba_kernel_function_regular,
            _np.array(operator_descriptor.options, dtype=data_type),
            grids_identical,
            dual_to_range.shapeset.evaluate,
            domain.shapeset.evaluate,
            result,
        )


def potential_assembler(device_interface, space, operator_descriptor, points, parameters):
    """Return an evaluator function to evaluate a potential."""
    from bempp_cl.core.numba_kernels import select_numba_kernels
    from bempp_cl.api.integration.triangle_gauss import rule
    from bempp_cl.api.utils.helpers import get_type

    (numba_assembly_function, numba_kernel_function_regular) = select_numba_kernels(
        operator_descriptor, mode="potential"
    )

    quad_points, quad_weights = rule(parameters.quadrature.regular)

    # Perform Numba assembly always in double precision
    # precision = operator_descriptor.precision
    precision = "double"

    dtype = _np.dtype(get_type(precision).real)

    if operator_descriptor.is_complex:
        result_type = _np.dtype(get_type(precision).complex)
    else:
        result_type = dtype

    kernel_dimension = operator_descriptor.kernel_dimension

    points_transformed = points.astype(dtype)
    grid_data = space.grid.data(precision)

    kernel_parameters = _np.array(operator_descriptor.options, dtype=dtype)

    def evaluator(x):
        """Actually evaluate the potential."""
        return numba_assembly_function(
            dtype,
            result_type,
            kernel_dimension,
            points_transformed,
            x.astype(result_type),
            grid_data,
            quad_points.astype(precision),
            quad_weights.astype(precision),
            space.number_of_shape_functions,
            space.shapeset.evaluate,
            numba_kernel_function_regular,
            kernel_parameters,
            space.normal_multipliers,
            space.support_elements,
        )

    return evaluator



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 15 16:07:10 2022

@author: tom
"""
import numpy as np

def SyncSineSweep(f1, f2, T, Fs, weight=0.8):
    """

    Parameters
    ----------
    f1 : start frequency.
    f2 : stop frequency.
    T : time length of sine sweep.
    Fs : Frequency sampling

    Returns
    -------
    t : time vector
    sss : synchronized sine sweep.

    """
    Ts = 1 / Fs
    t = np.arange(0, T, Ts)
    L = 1 / f1 * np.round(T * f1 / np.log(f2 / f1))
    syncSine = weight * np.sin(2 * np.pi * f1 * L * np.exp(t / L))
    return t, syncSine



"""Basic operator imports and definitions."""

import collections as _collections
import numpy as _np

from . import boundary
from . import potential
from . import far_field


OperatorDescriptor = _collections.namedtuple(
    "OperatorDescriptor",
    "identifier options kernel_type assembly_type precision is_complex singular_part kernel_dimension",
)

MultitraceOperatorDescriptor = _collections.namedtuple(
    "MultitraceOperatorDescriptor",
    "identifier options multitrace_kernel singular_contribution",
)


def _add_wavenumber(options, wavenumber, identifier="WAVENUMBER"):
    """Add a real/complex wavenumber to an options array."""
    if _np.iscomplexobj(wavenumber):
        options[identifier + "_REAL"] = 1.0 * _np.real(wavenumber)
        options[identifier + "_COMPLEX"] = 1.0 * _np.imag(wavenumber)
    else:
        options[identifier + "_REAL"] = 1.0 * _np.real(wavenumber)



"""Interface to DOLFIN for FEM-BEM coupling."""

import dolfin as _dolfin


def boundary_grid_from_fenics_mesh(fenics_mesh):
    """
    Create a Bempp boundary grid from a FEniCS Mesh.

    Return the Bempp grid and a map from the node numberings of the FEniCS
    mesh to the node numbers of the boundary grid.
    """
    import bempp_cl.api
    import numpy as np

    boundary_mesh = _dolfin.BoundaryMesh(fenics_mesh, "exterior", False)
    bm_coords = boundary_mesh.coordinates()
    bm_cells = boundary_mesh.cells()
    bm_nodes = boundary_mesh.entity_map(0).array().astype(np.int64)
    bempp_boundary_grid = bempp_cl.api.Grid(bm_coords.transpose(), bm_cells.transpose())
    return bempp_boundary_grid, bm_nodes


def fenics_to_bempp_trace_data(fenics_space):
    """Return tuple (space,trace_matrix)."""
    family, degree = fenics_space_info(fenics_space)

    if family == "Lagrange":
        if degree == 1:
            return p1_trace(fenics_space)
        else:
            raise NotImplementedError()
    elif family == "Nedelec 1st kind H(curl)":
        if degree == 1:
            return nc1_tangential_trace(fenics_space)
        else:
            raise NotImplementedError()
    else:
        raise NotImplementedError()


def fenics_space_info(fenics_space):
    """Return tuple (family,degree) containing information about a FEniCS space."""
    element = fenics_space.ufl_element()
    family = element.family()
    degree = element.degree()
    return (family, degree)


class FenicsOperator(object):
    """Wrap a FEniCS Operator into a Bempp operator."""

    def __init__(self, fenics_weak_form):
        """Construct an operator from a weak form in FEniCS."""
        self._fenics_weak_form = fenics_weak_form
        self._sparse_mat = None

    def weak_form(self):
        """Return the weak form."""
        from bempp_cl.api.assembly.discrete_boundary_operator import (
            SparseDiscreteBoundaryOperator,
        )
        from dolfin import as_backend_type, assemble, parameters
        from scipy.sparse import csr_matrix

        if self._sparse_mat is not None:
            return SparseDiscreteBoundaryOperator(self._sparse_mat)

        backend = parameters["linear_algebra_backend"]
        if backend != "PETSc":
            raise ValueError("Only the PETSc linear algebra backend is supported.")
        mat = as_backend_type(assemble(self._fenics_weak_form)).mat()
        (indptr, indices, data) = mat.getValuesCSR()
        self._sparse_mat = csr_matrix((data, indices, indptr), shape=mat.size)

        return SparseDiscreteBoundaryOperator(self._sparse_mat)


# pylint: disable=too-many-locals
def p1_trace(fenics_space):
    """
    Return the P1 trace operator.

    This function returns a pair (space, trace_matrix),
    where space is a Bempp space object and trace_matrix is the corresponding
    matrix that maps the coefficients of a FEniCS function to its boundary
    trace coefficients in the corresponding Bempp space.

    """
    import bempp_cl.api
    from scipy.sparse import coo_matrix
    import numpy as np

    family, degree = fenics_space_info(fenics_space)
    if not (family == "Lagrange" and degree == 1):
        raise ValueError("fenics_space must be a p1 Lagrange space")

    fenics_mesh = fenics_space.mesh()
    bempp_boundary_grid, bm_nodes = boundary_grid_from_fenics_mesh(fenics_mesh)

    # First get trace space
    space = bempp_cl.api.function_space(bempp_boundary_grid, "P", 1)

    # FEniCS vertices to boundary dofs
    b_vertices_from_vertices = coo_matrix(
        (np.ones(len(bm_nodes)), (np.arange(len(bm_nodes)), bm_nodes)),
        shape=(len(bm_nodes), fenics_mesh.num_vertices()),
        dtype="float64",
    ).tocsc()

    # Finally FEniCS dofs to vertices.
    vertices_from_fenics_dofs = coo_matrix(
        (
            np.ones(fenics_mesh.num_vertices()),
            (
                _dolfin.dof_to_vertex_map(fenics_space),
                np.arange(fenics_mesh.num_vertices()),
            ),
        ),
        shape=(fenics_mesh.num_vertices(), fenics_mesh.num_vertices()),
        dtype="float64",
    ).tocsc()

    # Get trace matrix by multiplication
    trace_matrix = b_vertices_from_vertices @ vertices_from_fenics_dofs

    # Now return everything
    return space, trace_matrix


# pylint: disable=too-many-locals
def nc1_tangential_trace(fenics_space):
    """
    Return the NC1 (twisted) tangential trace operator.

    This function returns a pair (space, trace_matrix),
    where space is a Bempp space object and trace_matrix is the corresponding
    matrix that maps the coefficients of a FEniCS function to its boundary
    trace coefficients in the corresponding Bempp space.

    """
    import bempp_cl.api
    from scipy.sparse import coo_matrix
    import numpy as np

    family, degree = fenics_space_info(fenics_space)
    if not (family == "Nedelec 1st kind H(curl)" and degree == 1):
        raise ValueError("fenics_space must be an order 1 Nedelec 1st kind H(curl) space")

    fenics_mesh = fenics_space.mesh()
    bempp_boundary_grid, bm_nodes = boundary_grid_from_fenics_mesh(fenics_mesh)
    fenics_dim = fenics_space.dim()

    # First get trace space
    space = bempp_cl.api.function_space(bempp_boundary_grid, "RWG", 0)

    # Now compute the mapping from Bempp dofs to FEniCS dofs
    # Overall plan:
    #   bempp_dofs <- bd_vertex_pairs <- all_vertex_pairs <- all_edges <- fenics_dofs

    # First the Bempp dofs to the boundary edges
    #   bempp_dofs <- bd_vertex_pairs

    grid = space.grid
    edge_count = space.global_dof_count
    face_count = grid.entity_count(0)

    dof_to_vertices_map = np.zeros((edge_count, 2), dtype=np.int64)
    dof_to_face_map = [None] * edge_count

    for element in grid.entity_iterator(0):
        g_d = space.local2global[element.index]
        vs = [v.index for v in element.sub_entity_iterator(2)]
        for i, e in enumerate([(0, 1), (0, 2), (1, 2)]):
            dof_to_vertices_map[g_d[i], 0] = vs[e[0]]
            dof_to_vertices_map[g_d[i], 1] = vs[e[1]]

            dof_to_face_map[g_d[i]] = (element, i)

    # Now the boundary triangle edges to the tetrahedron faces
    #   bd_vertex_pairs <- all_vertex_pairs
    # is bm_nodes

    #   all_vertex_pairs <- all_edges
    all_vertices_to_all_edges = {}
    for edge in _dolfin.entities(fenics_mesh, 1):
        i = edge.index()
        ent = edge.entities(0)
        all_vertices_to_all_edges[(max(ent), min(ent))] = i

    # Make matrix bempp_dofs <- all_edges
    bempp_dofs_to_all_edges_map = np.zeros(space.global_dof_count, dtype=np.int64)
    for bd_e, v in enumerate(dof_to_vertices_map):
        v2 = bm_nodes[v]
        v2 = (max(v2), min(v2))
        all_e = all_vertices_to_all_edges[v2]
        bempp_dofs_to_all_edges_map[bd_e] = all_e

    bempp_dofs_from_all_edges = coo_matrix(
        (
            np.ones(space.global_dof_count),
            (np.arange(space.global_dof_count), bempp_dofs_to_all_edges_map),
        ),
        shape=(space.global_dof_count, fenics_dim),
        dtype="float64",
    ).tocsc()

    # Finally the edges to FEniCS dofs
    #   all_edges <- fenics_dofs
    dofmap = fenics_space.dofmap()
    dof_to_edge_map = np.zeros(fenics_dim, dtype=np.int64)
    for cell in _dolfin.cells(fenics_mesh):
        c_d = dofmap.cell_dofs(cell.index())
        c_e = cell.entities(1)
        for d, e in zip(c_d, c_e):
            dof_to_edge_map[d] = e

    all_edges_from_fenics_dofs = coo_matrix(
        (np.ones(fenics_dim), (dof_to_edge_map, np.arange(fenics_dim))),
        shape=(fenics_dim, fenics_dim),
        dtype="float64",
    ).tocsc()

    # Get trace matrix by multiplication
    #   bempp_dofs <- bd_edges <- all_edges <- fenics_dofs
    trace_matrix = bempp_dofs_from_all_edges * all_edges_from_fenics_dofs

    # Now sort out directions
    v_list = list(space.grid.entity_iterator(2))
    local_coords = [
        np.array([[0.5], [0.0]]),
        np.array([[0.0], [0.5]]),
        np.array([[0.5], [0.5]]),
    ]

    # Build a map from Bempp triangles to FEniCS edges
    # aim: bempp_triangles -> all_edge_triplets -> fenics_tetrahedra
    # First: bempp Triangles -> all edge triplets

    triangles_to_edge_triples = [None] * face_count
    for face in grid.entity_iterator(0):
        es = [e.index for e in face.sub_entity_iterator(2)]
        all_es = [bm_nodes[e] for e in es]
        all_es.sort()
        triangles_to_edge_triples[face.index] = tuple(all_es)
    edge_triples_to_tetrahedra = {}
    for tetra in _dolfin.cells(fenics_mesh):
        all_es = [v.index() for v in _dolfin.vertices(tetra)]
        all_es.sort()
        edge_triples_to_tetrahedra[(all_es[0], all_es[1], all_es[2])] = tetra
        edge_triples_to_tetrahedra[(all_es[0], all_es[1], all_es[3])] = tetra
        edge_triples_to_tetrahedra[(all_es[0], all_es[2], all_es[3])] = tetra
        edge_triples_to_tetrahedra[(all_es[1], all_es[2], all_es[3])] = tetra
    triangles_to_tetrahedra = [edge_triples_to_tetrahedra[triple] for triple in triangles_to_edge_triples]

    # return None, None
    fenics_fun = _dolfin.Function(fenics_space)
    non_z = trace_matrix.nonzero()
    for i, j in zip(non_z[0], non_z[1]):
        fenics_fun.vector()[:] = [1.0 if k == j else 0.0 for k in range(fenics_dim)]
        bempp_fun = bempp_cl.api.GridFunction(
            space,
            coefficients=[1.0 if k == i else 0.0 for k in range(space.global_dof_count)],
        )

        v1 = v_list[dof_to_vertices_map[i][0]]
        v2 = v_list[dof_to_vertices_map[i][1]]
        midpoint = (v1.geometry.corners + v2.geometry.corners) / 2

        face = dof_to_face_map[i]
        fenics_values = np.zeros(3)
        fenics_fun.eval_cell(fenics_values, midpoint.T[0], triangles_to_tetrahedra[face[0].index])

        bempp_values = bempp_fun.evaluate(face[0].index, local_coords[face[1]])
        normal = face[0].geometry.normal
        cross = np.cross(fenics_values, normal)
        k = np.argmax(np.abs(cross))

        trace_matrix[i, j] = cross[k] / bempp_values[k, 0]

    # Now return everything
    return space, trace_matrix



import generalToolbox as gtb
import matplotlib.pyplot as plt
import numpy as np
from numpy import cos, sin, tan, exp, pi
import scipy as sp
from scipy.fft import fft
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit



class impedance:
    def __init__(self, impedance_file, Bl, peak_max_bound=3000, skiprows=1, delimiter=",",
                 freqCol=0, HCol=1, PCol=2, usecols=None, decimate=1):
        self.impedance_file = impedance_file
        self.peak_max_bound = peak_max_bound


        # 1. load data
        data = np.loadtxt(impedance_file, skiprows=skiprows, delimiter=delimiter, usecols=usecols)
        self.freq = data[::decimate, freqCol]
        self.finterp = np.arange(self.freq[0], self.freq[-1] + 0.1, 0.1)
        try:
            self.Ze_meas = data[::decimate, HCol] * np.exp(1j * data[::decimate, PCol])  # H*e^(j*Phi)
        except:
            self.Ze_meas = data[::decimate, HCol]

        # Run a first optimization to fit electrical parameters
        # and some mechanical parameters
        self.params = self.get_first_opt()
        self.Zin = get_Zin(self.finterp, self.params['Re'], self.params['Le'],
                           self.params['Res'], self.params['R2'], self.params['L2'],
                           self.params['R3'], self.params['L3'], self.params['Ws'],
                           self.params['Qms'])

        # Run a second optimization to fit remaining mechanical parameters
        s = 2j * np.pi * self.finterp
        self.Zm = self.params['Res'] / (1 + self.params['Qms'] * (s / self.params['Ws'] + self.params['Ws'] / s))
        Mms, Rms, Cms = estimate_pmec(Bl, self.params['Qms'], self.params['Res'], self.params['fs'])
        self.params['Mms'] = Mms
        self.params['Cms'] = Cms
        self.params['Rms'] = Rms
        self.params['Bl'] = Bl
        # self.get_mec_opt()



    def get_first_opt(self):
        # 2. Get some visible values
        ind_bound, _ = gtb.findInArray(self.freq, self.peak_max_bound)
        Zmax = np.max(np.abs(self.Ze_meas[:ind_bound]))
        ind_fc, _ = gtb.findInArray(np.abs(self.Ze_meas), Zmax)
        fc = self.freq[ind_fc]

        # 3. Reduce the size of frequency range if it goes above 20 kHz
        ind_fmax, fmax = gtb.findInArray(self.freq, 20e3)
        freq = self.freq[:ind_fmax + 1]
        om = 2 * np.pi * freq
        s = 1j * om
        Ze_meas = self.Ze_meas[:ind_fmax + 1]

        # 4. Start Values
        Re = np.min(np.abs(self.Ze_meas))
        Le = 1e-5
        Res = Zmax - Re
        R2, L2 = 0.5, 2e-4
        R3, L3 = 0.25, 1e-4
        Wc, Qmc = 2 * np.pi * fc, getQMC(freq, Ze_meas)
        x0 = [Re, Le, Res, R2, L2, R3, L3, Wc, Qmc]  # initial parameters

        # 5. Merci Scipy
        popt, pcov = sp.optimize.curve_fit(sealed_impedance_calc,
                                           freq, np.abs(Ze_meas), p0=x0, maxfev=16000)
        Re, Le = popt[0], popt[1]
        Res = popt[2]
        R2, L2 = popt[3], popt[4]
        R3, L3 = popt[5], popt[6]
        Wc, Qmc = popt[7], popt[8]
        fc = Wc / 2 / np.pi
        sc = 1j * Wc

        # impedance at resonance
        Zfc = np.abs(Re + sc * Le + gtb.parallel(sc * L2, R2) + gtb.parallel(sc * L3, R3))

        # # Q factors
        # Qec = np.sqrt(Mmc / Cmc) * (Zfc / Bl ** 2)
        # Qtc = Qmc * Qec / (Qmc + Qec)

        # out param
        params = {}
        params['Re'] = Re
        params['Le'] = Le
        params['Res'] = Res
        params['R2'] = R2
        params['L2'] = L2
        params['R3'] = R3
        params['L3'] = L3
        params['Ws'] = Wc
        params['fs'] = fc
        params['Qms'] = Qmc
        params['Zfs'] = Zfc
        # params['Cmc'] = Cmc
        # params['Mmc'] = Mmc
        # params['Zfc'] = Zfc
        # params['Qec'] = Qec
        # params['Qtc'] = Qtc
        return params

    # def get_mec_opt(self):
    #     Bl_est = estimate_Bl(self.finterp, self.Zm, self.Zin,
    #                          self.params['Res'], self.peak_max_bound)
    #     # print('estimated Bl: ', Bl_est)
    #     Mms_est, Rms_est, Cms_est = estimate_pmec(Bl_est, self.params['Qms'],
    #                                               self.params['Res'],
    #                                               self.params['fs'])
    #     # print('estimated Mms: ', Mms_est)
    #     # print('estimated Rms: ', Rms_est)
    #     # print('estimated Cms: ', Cms_est)
    #
    #     x1 = [Bl_est, Rms_est, Mms_est, Cms_est]
    #     Zems = self.Zin - (2j*np.pi*self.finterp * self.params['Le'] + self.params['Re'])
    #     popt2, pcov2 = sp.optimize.curve_fit(Zes_impedance, self.finterp, np.abs(self.Zm), p0=x1, maxfev=100000)
    #     self.params['Bl'] = popt2[0]
    #     self.params['Rms'] = popt2[1]
    #     self.params['Mms'] = popt2[2]
    #     self.params['Cms'] = popt2[3]
    #     return None

    def get_Zin(self, frequency):
        s = 2j*np.pi*frequency
        Zms = self.params['Rms'] + s*self.params['Mms'] + 1/s/self.params['Cms']
        Zm = self.params['Bl']**2 / Zms
        Ze = (self.params['Re'] + s*self.params['Le'] +
              gtb.parallel(self.params['R2'], s*self.params['L2']) +
              gtb.parallel(self.params['R3'], s*self.params['L3']))
        Zin = Ze+Zm
        return Zin

    def plot_Zin(self, measured=False, save=False):
        if measured is True:
            Z = self.Ze_meas
            freq = self.freq
            title='measured_impedance'
        elif measured is False:
            Z = self.Zin
            freq = self.finterp
            title = 'fitted_impedance'
        elif measured == "both":
            Z1 = self.Ze_meas
            freq_1 = self.freq
            Z2 = self.Zin
            freq_2 = self.finterp
            title = 'measured_fitted_impedance'

        if measured is True or measured is False:
            fig, ax = plt.subplots()
            ax.semilogx(freq, np.abs(Z), 'b',
                        label="modulus")
            ax.legend(loc='upper left')
            ax.grid(linestyle='dotted', which='both')
            ax.set(ylabel="|Z| [Ohm]", xlabel="Frequency [Hz]")
            ax2 = ax.twinx()
            ax2.semilogx(freq, np.angle(Z), ':r',
                         label="phase")
            ax2.legend(loc='upper right')
            ax2.set(ylabel="phase [rads]", ylim=[-pi / 2, pi / 2])
            ax2.grid()
            plt.tight_layout()
            if save is True:
                plt.savefig(title, dpi=256)

        elif measured == 'both':
            fig, ax = plt.subplots()
            ax.semilogx(freq_1, np.abs(Z1), 'b',
                        label="modulus - measured")
            ax.semilogx(freq_2, np.abs(Z2), 'g',
                        label="modulus - fitted")
            ax.legend(loc='upper left')
            ax.grid(linestyle='dotted', which='both')
            ax.set(ylabel="|Z| [Ohm]", xlabel="Frequency [Hz]")
            ax2 = ax.twinx()
            ax2.semilogx(freq_1, np.angle(Z1), ':r',
                         label="phase - measured", alpha=0.45)
            ax2.semilogx(freq_2, np.angle(Z2), ':', color='black',
                         label="phase - fitted")
            ax2.legend(loc='upper right')
            ax2.set(ylabel="phase [rads]", ylim=[-pi / 2, pi / 2])
            ax2.grid()
            plt.tight_layout()
            if save is True:
                plt.savefig(title, dpi=256)
        return None

def sealed_impedance_calc(freq, Re, Le, Res, R2, L2, R3, L3, Wc, Qmc):
    """
    sealed impedance calculator, should be use within the curve_fit() algorithm
    :param freq:
    :param Re:
    :param Le:
    :param Res:
    :param R2:
    :param L2:
    :param R3:
    :param L3:
    :param Wc:
    :param Qmc:
    :return:
    """
    s = 1j * 2 * np.pi * freq

    # mechanical impedance
    fncZm = Res / (1 + Qmc * (s / Wc + Wc / s))

    # electrical impedance
    fncZe = Re + s * Le + gtb.parallel(s * L2, R2) + gtb.parallel(s * L3, R3)

    # total impedance
    fncZ = fncZe + fncZm
    return np.abs(fncZ)

def get_Zin(freq, Re, Le, Res, R2, L2, R3, L3, Wc, Qmc):
    """
    sealed impedance calculator, should be use within the curve_fit() algorithm
    :param freq:
    :param Re:
    :param Le:
    :param Res:
    :param R2:
    :param L2:
    :param R3:
    :param L3:
    :param Wc:
    :param Qmc:
    :return:
    """
    s = 1j * 2 * np.pi * freq

    # mechanical impedance
    fncZm = Res / (1 + Qmc * (s / Wc + Wc / s))

    # electrical impedance
    fncZe = Re + s * Le + gtb.parallel(s * L2, R2) + gtb.parallel(s * L3, R3)

    # total impedance
    fncZ = fncZe + fncZm
    return fncZ

def getQMC(freq, Z, stepf=1e-3):
    """
    Find the quality factor of the impedance peak of a driver in a sealed box.
    :param freq:
    :param Z:
    :param stepf:
    :return:
    """
    # increase a bit the resolution (doesn't change curve's shape)
    finterp = np.arange(freq[0], freq[-1] + stepf, stepf)
    Zinterp = np.interp(finterp, freq, np.abs(Z))

    # get index of Zmax (index of f0) and Zmax
    ind_f0, Zmax = gtb.findInArray(np.abs(Zinterp), np.max(np.abs(Zinterp)))
    Ef0 = finterp[ind_f0]

    ind_fmin, Zfmin = gtb.findInArray(np.abs(Zinterp[:ind_f0]),
                                      np.max(np.abs(Zinterp)) * 0.707)  # get -3 dB value before resonance
    ind_fmax, Zfmax = gtb.findInArray(np.abs(Zinterp[ind_f0:]), np.max(np.abs(Zinterp)) * 0.707)
    range = [finterp[ind_fmin], finterp[ind_f0 + ind_fmax]]
    Qs = finterp[ind_f0] / (range[1] - range[0])
    return Qs

def Zes_impedance(freq, Bl, Rms, Mms, Cms):
    s = 2j * np.pi * freq
    Zems = Bl ** 2 / (Rms + s*Mms + 1/s/Cms)
    return np.abs(Zems)

def estimate_pmec(Bl, Qms, Res, fs):
    Mms = Qms * Bl**2 / 2 / np.pi / fs / Res
    Rms = Bl**2 / Res
    Cms = Res**2 / Bl**4 * Mms / Qms**2
    return Mms, Rms, Cms

def estimate_Bl(freq, Zm, Zin, Res, max_bound):
    ind_max, _ = gtb.findInArray(freq, max_bound)
    freq = freq[:ind_max]
    Zm = Zm[:ind_max]
    Zin = Zin[:ind_max]
    max_Zm = np.max(np.abs(Zm))
    max_Zin = np.max(np.abs(Zin))

    Bl_est = 0.01
    Zfc = max_Zin - max_Zm / Bl_est**2
    while Zfc < Res:
        Bl_est += 0.01
        Zfc = max_Zin - max_Zm / Bl_est
    return Bl_est


"""
Functions to manipulate frequency related problems (decimate, octave smoothing)
"""

import numpy as np
from tqdm import tqdm
import pyuff

## manipulate frequency array (Hz)
def laplace(f):
    """
    Convert frequency values from Hertz to complex frequency in the Laplace domain.

    Parameters
    ----------
    f : numpy array
        Frequency axis in Hertz.

    Returns
    -------
    s : numpy array
        Complex frequency values corresponding to the Laplace domain (2*pi*freq*1j).

    Notes
    -----
    This function takes an array of frequency values in Hertz and converts them to
    complex frequency values in the Laplace domain using the formula 2*pi*freq*1j,
    where 'freq' is the input frequency values. The resulting 's' values can be used
    for Laplace domain analysis of systems.
    """
    return 2 * np.pi * f * 1j


def octave_smoothing(X_FFT, freq_data, nthOctBand):
    """
    Apply octave band smoothing to frequency response data.

    Parameters
    ----------
    X_FFT : numpy array
        Dataset to be smoothed (Frequency Response in SPL).

    freq_data : numpy array
        Frequency axis corresponding to the data.

    nthOctBand : int
        Order of the octave band to use for filtering.

    Returns
    -------
    smoothed_data : numpy array
        Frequency response data after applying octave band smoothing.

    Notes
    -----
    This function applies octave band smoothing to the input frequency response data
    'X_FFT'. The smoothing is performed using the specified 'nthOctBand' order of
    octave band filtering. The 'freq_data' parameter is the frequency axis corresponding
    to the data. The function calculates and applies the smoothing coefficients based on
    the octave band filtering approach, resulting in smoothed frequency response data.
    """
    f_min = np.argwhere(freq_data > 1)[0][0]  # first value superior to zero
    x_fft = np.zeros(len(X_FFT))  # data to be smoothed
    if nthOctBand > 0:
        print('Filtering Data, 1/{} octave band'.format(nthOctBand))
        for i in tqdm(np.arange(f_min, len(freq_data))):
            fv = freq_data[i]
            s = (fv / nthOctBand) / np.pi
            g = np.exp(-(((freq_data - fv) ** 2) / (2 * (s ** 2))))
            g_s = g / np.sum(g)
            x_fft[i] = np.sum(g_s * X_FFT)  # smoothed coefficients
        if X_FFT.all() >= 0:
            x_fft[x_fft < 0] = 0  # set all negative data to zero
    return x_fft

def smooth_directivity(X_FFT, freq_data, nthOctBand):
    """
    Apply octave band smoothing to frequency response data.

    Parameters
    ----------
    X_FFT : numpy array
        Dataset to be smoothed (Frequency Response in SPL).

    freq_data : numpy array
        Frequency axis corresponding to the data.

    nthOctBand : int
        Order of the octave band to use for filtering.

    Returns
    -------
    smoothed_data : numpy array
        Frequency response data after applying octave band smoothing.

    Notes
    -----
    This function applies octave band smoothing to the input frequency response data
    'X_FFT'. The smoothing is performed using the specified 'nthOctBand' order of
    octave band filtering. The 'freq_data' parameter is the frequency axis corresponding
    to the data. The function calculates and applies the smoothing coefficients based on
    the octave band filtering approach, resulting in smoothed frequency response data.
    """
    f_min = np.argwhere(freq_data > 1)[0][0]  # first value superior to zero
    x_fft = np.zeros(np.shape(X_FFT))  # data to be smoothed
    if nthOctBand > 0:
        print('Filtering Data, 1/{} octave band'.format(nthOctBand))
        for i in tqdm(np.arange(f_min, len(freq_data))):
            fv = freq_data[i]
            s = (fv / nthOctBand) / np.pi
            g = np.exp(-(((freq_data - fv) ** 2) / (2 * (s ** 2))))
            g_s = g / np.sum(g)
            for j in range(len(X_FFT)):
                x_fft[j, i] = np.sum(g_s * X_FFT[j, :])  # smoothed coefficients
        if X_FFT.all() >= 0:
            x_fft[x_fft < 0] = 0  # set all negative data to zero
    return x_fft

def decimate_frequency_axis(old_axis, new_axis):
    """
    Decimate an existing frequency axis into a new one by finding the closest values.

    Parameters:
    old_axis (numpy.ndarray): The existing frequency axis from which values will be selected.
    new_axis (numpy.ndarray): The new frequency axis to decimate onto.

    Returns:
    numpy.ndarray: The decimated frequency axis containing values 
    from the old_axis that are closest to the values in new_axis.

    Example:
        >>> old_axis = np.linspace(0, 100, 1000)
        >>> new_axis = np.logspace(1, 2, 50)
        >>> decimated_result = decimate_frequency_axis(old_axis, new_axis)
        >>> print(decimated_result)
    """
    decimated_axis = np.empty_like(new_axis)
    closest_idx = np.zeros(len(new_axis), dtype=int)

    for i, new_val in enumerate(new_axis):
        closest_idx_tmp = np.argmin(np.abs(old_axis - new_val))
        decimated_axis[i] = old_axis[closest_idx_tmp]
        closest_idx[i] = int(closest_idx_tmp)

    return decimated_axis, closest_idx


## create frequency arrays
def freq_log10(start, stop, numpoints, Nd=2):
    """
    Generate a logarithmically spaced frequency array.

    Parameters
    ----------
    start : float
        Starting frequency (lowest frequency) in Hertz.

    stop : float
        Stopping frequency (highest frequency) in Hertz.

    numpoints : int
        Number of points in the generated frequency array.
    
    Nd : int
        Number of decimals

    Returns
    -------
    freq_array : numpy array
        Logarithmically spaced frequency array spanning from 'start' to 'stop'.

    Notes
    -----
    This function generates a numpy array of 'numpoints' logarithmically spaced
    frequency values between 'start' and 'stop'. The frequency values are evenly
    distributed on a logarithmic scale, making them suitable for logarithmic plots
    or analyses where the frequency range spans multiple orders of magnitude.
    """
    freq_array = np.logspace(np.log10(start), np.log10(stop), numpoints)
    return np.round(freq_array, Nd)


def freq_uff(uff_file, freq_axis_user):
    """
    Create frequency axis from uff dataset. Will return closest frequencies to freq_axis_user
    :param uff_file:
    :param freq_axis_wanted:
    :return:
    """
    uff_data          = pyuff.UFF(uff_file)
    data              = uff_data.read_sets()
    freq_axis_meas    = data[-1]['x']
    freq_out, ind_out = decimate_frequency_axis(freq_axis_meas, freq_axis_user)
    return np.unique(freq_out)



"""Duffy type rules for collocation integrals."""

import numpy as _np


def duffy_rule_on_reference_triangle(order):
    """Return points and weights for Duffy rule on unit triangle.

    The triangle has coordinates (0, 0), (1, 0), and (1, 1). The
    singularity is assumed to be at (0, 0).
    """
    from bempp_cl.api.integration.gauss import rule

    gauss_points, gauss_weights = rule(order)
    npoints = len(gauss_weights)

    points = _np.empty((2, npoints * npoints), dtype=_np.float64)
    weights = _np.empty(npoints * npoints, dtype=_np.float64)

    count = 0
    for index1 in range(npoints):
        for index2 in range(npoints):
            points[:, count] = [
                gauss_points[index1],
                gauss_points[index1] * gauss_points[index2],
            ]
            weights[count] = gauss_weights[index1] * gauss_weights[index2] * gauss_points[index1]
            count += 1
    return points, weights


def singular_collocation_rule_piecewise_const(order):
    """Singular collocation integral for one singularity on unit triangle barycenter."""
    duffy_points, duffy_weights = duffy_rule_on_reference_triangle(order)
    npoints = len(duffy_weights)

    points = _np.empty((2, 3 * npoints), dtype=_np.float64)
    weights = _np.empty(3 * npoints, dtype=_np.float64)

    triangle_points = [
        _np.array([[1.0 / 3, 1.0 / 3], [0.0, 0], [1.0, 0]]).T,
        _np.array([[1.0 / 3, 1.0 / 3], [1, 0], [0, 1]]).T,
        _np.array([[1.0 / 3, 1.0 / 3], [0, 1], [0, 0]]).T,
    ]

    for index in range(3):
        v0 = triangle_points[index][:, 0].reshape(2, 1)
        A = _np.hstack(
            [
                (triangle_points[index][:, 1] - triangle_points[index][:, 0]).reshape(2, 1),
                (triangle_points[index][:, 2] - triangle_points[index][:, 1]).reshape(2, 1),
            ]
        )
        points[:, index * npoints : (1 + index) * npoints] = v0 + A @ duffy_points
        weights[index * npoints : (1 + index) * npoints] = _np.abs(_np.linalg.det(A)) * duffy_weights

    return points, weights



"""
Collection of electro-dynamic circuit blocks 

"""
import random, string

def randomblock_id(length):
   letters = string.ascii_lowercase
   return ''.join(random.choice(letters) for i in range(length))

class EAD:
    def __init__(self, A, B, C, D, 
                 Le, Re, Cms, Mms, Rms, Bl, Sd, v_probe=None):
        """
        Creates an electro-acoustic driver based on its input/output nodes and
        T/S parameters.

        Parameters
        ----------
        A : int or str
            positive electrical node.
        B : int or str
            negative electrical node..
        C : int or str
            positive acoustical node..
        D : int or str
            negative acoustical node..
        Le : float
            coil inductance.
        Re : float
            coil's electrical resistance.
        Cms : float
            suspensions compliance.
        Mms : float
            moving mass.
        Rms : float
            mechanical losses.
        Bl : float
            electrical to mechanical force factor.
        Sd : float
            effective radiating surface.

        Returns
        -------
        None.

        """
        from electroacPy.circuitSolver.components.electric import resistance, inductance, capacitance
        from electroacPy.circuitSolver.components.coupler import CCVS 
        
        
        self.Re = Re
        self.Le = Le 
        self.Bl = Bl
        self.Mms = Mms
        self.Cms = Cms
        self.Rms = Rms
        self.Sd = Sd
        
        np = str(A)   # input electric
        nm = str(B)   # output electric
        np1 = str(C)  # input acoustic
        nm1 = str(D)  # output acoustic
        rnd_id = randomblock_id(3)
        
        if v_probe is not None:
            v_p = v_probe
        elif v_probe is None:
            v_p = np + "_m_8_" + rnd_id

        self.network = {"Le": inductance(np, np+"_e1_"+rnd_id, self.Le),
                        "Re": resistance(np+"_e1_"+rnd_id, np+"_e2_"+rnd_id, self.Re),
                        "Bl1": CCVS(np+"_e2_"+rnd_id, np+"_e3_"+rnd_id,
                                    np+"_m4_"+rnd_id, np+"_m5_"+rnd_id, self.Bl),
                        "Bl2": CCVS(np+"_m5_"+rnd_id, 0, 
                                    np+"_e3_"+rnd_id, nm, -self.Bl),
                        "Mms": inductance(np+"_m4_"+rnd_id, 
                                          np+"_m6_"+rnd_id, self.Mms),
                        "Cms": capacitance(np+"_m6_"+rnd_id, 
                                           np+"_m7_"+rnd_id, self.Cms),
                        "Rms": resistance(np+"_m7_"+rnd_id, 
                                          v_p, self.Rms),
                        "Sd1": CCVS(v_p, np+"_m9_"+rnd_id, 
                                    np+"_s10_"+rnd_id, np+"_s11_"+rnd_id, self.Sd),
                        "RSn": resistance(np+"_s10_"+rnd_id, np+"_s12_"+rnd_id, 1e-9),
                        "Sd2": CCVS(np+"_s12_"+rnd_id, np+"_s13_"+rnd_id, 
                                    np1, np1+"_a14_"+rnd_id, 1),
                        "Sd3": CCVS(np1+"_a14_"+rnd_id, nm1, 
                                    np+"_s13_"+rnd_id, 0, -1),
                        "Sd4": CCVS(np+"_s11_"+rnd_id, 0, 
                                    np+"_m9_"+rnd_id, 0, -self.Sd)
                        }     
        
        
class EADImport:
    def __init__(self, A, B, C, D, lpmFile, v_probe=None, rho=1.22, c=343):
        """
        Creates an electro-acoustic driver based on its input/output nodes and
        T/S parameters --- from LPM data-file.

        Parameters
        ----------
        A : int or str
            positive electrical node.
        B : int or str
            negative electrical node..
        C : int or str
            positive acoustical node..
        D : int or str
            negative acoustical node..
        Le : float
            coil inductance.
        Re : float
            coil's electrical resistance.
        Cms : float
            suspensions compliance.
        Mms : float
            moving mass.
        Rms : float
            mechanical losses.
        Bl : float
            electrical to mechanical force factor.
        Sd : float
            effective radiating surface.

        Returns
        -------
        None.

        """
        from electroacPy.circuitSolver.components.electric import resistance, inductance, capacitance
        from electroacPy.circuitSolver.components.coupler import CCVS 
        from electroacPy.speakerSim.electroAcousticDriver import loadLPM
        from numpy import arange
        
        drv = loadLPM(lpmFile, arange(10, 100, 1), c=c, rho=rho)
        
        self.Re = drv.Re
        self.Le = drv.Le 
        self.Bl = drv.Bl
        self.Mms = drv.Mms
        self.Cms = drv.Cms
        self.Rms = drv.Rms
        self.Sd = drv.Sd
        self.Vas = drv.Vas
        self.Fs = drv.Fs
        
        np = str(A)   # input electric
        nm = str(B)   # output electric
        np1 = str(C)  # input acoustic
        nm1 = str(D)  # output acoustic
        rnd_id = randomblock_id(3)
        
        if v_probe is not None:
            v_p = v_probe
        elif v_probe is None:
            v_p = np + "_m_8_" + rnd_id

        self.network = {"Le": inductance(np, np+"_e1_"+rnd_id, self.Le),
                        "Re": resistance(np+"_e1_"+rnd_id, np+"_e2_"+rnd_id, self.Re),
                        "Bl1": CCVS(np+"_e2_"+rnd_id, np+"_e3_"+rnd_id,
                                    np+"_m4_"+rnd_id, np+"_m5_"+rnd_id, self.Bl),
                        "Bl2": CCVS(np+"_m5_"+rnd_id, 0, 
                                    np+"_e3_"+rnd_id, nm, -self.Bl),
                        "Mms": inductance(np+"_m4_"+rnd_id, 
                                          np+"_m6_"+rnd_id, self.Mms),
                        "Cms": capacitance(np+"_m6_"+rnd_id, 
                                           np+"_m7_"+rnd_id, self.Cms),
                        "Rms": resistance(np+"_m7_"+rnd_id, 
                                          v_p, self.Rms),
                        "Sd1": CCVS(v_p, np+"_m9_"+rnd_id, 
                                    np+"_s10_"+rnd_id, np+"_s11_"+rnd_id, self.Sd),
                        "RSn": resistance(np+"_s10_"+rnd_id, np+"_s12_"+rnd_id, 1e-9),
                        "Sd2": CCVS(np+"_s12_"+rnd_id, np+"_s13_"+rnd_id, 
                                    np1, np1+"_a14_"+rnd_id, 1),
                        "Sd3": CCVS(np1+"_a14_"+rnd_id, nm1, 
                                    np+"_s13_"+rnd_id, 0, -1),
                        "Sd4": CCVS(np+"_s11_"+rnd_id, 0, 
                                    np+"_m9_"+rnd_id, 0, -self.Sd)
                        }     
        
        
        
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 17 08:29:39 2024

@author: tom
"""

import numpy as np
import pyvista
import os
import electroacPy
from ..general import plot as gplot
directory_path = os.path.abspath(electroacPy.__file__)

pi = np.pi

class evaluations:
    """
    Tools for creating and managing various types of acoustic evaluations 
    using a Boundary Element Method (BEM) object.

    Available evaluation types:
    - Planar evaluations: Defined on a planar surface with specified dimensions.
    - Polar evaluations: Defined along a circular path in a specified plane.
    - Spherical evaluations: Defined on a spherical surface.
    - Pressure response evaluations: Recorded pressure responses at specified microphone positions.
    - Custom plotting grids through import
    
    """
    def __init__(self, bemObject):
        self.bemObject = bemObject
        self.frequency = bemObject.frequency
        self.setup = {}
        
        # ref to system
        self.referenceStudy = None
    
    
    def polarRadiation(self, evaluationName, minAngle: float, maxAngle: float,
                       step: float, on_axis: str, direction: str, 
                       radius: float = 5, offset: list = [0, 0, 0], **kwargs):
        
        if evaluationName not in self.setup:
            self.setup[evaluationName] = PolarRadiation(minAngle, maxAngle, 
                                                         step, on_axis, direction,
                                                         radius, offset)
        elif evaluationName in self.setup and "overwrite" in kwargs:
            self.setup.pop(evaluationName)
            self.setup[evaluationName] = PolarRadiation(minAngle, maxAngle, 
                                                         step, on_axis, direction,
                                                         radius, offset)
        else:
             print("evaluation {} already exists. You can overwrite it using  \
                   the 'overwrite' flag".format(evaluationName))   
            
    
    def pressureField(self, evaluationName, Length: float, Width: float,
                      step: float, plane: str, offset: list = [0, 0, 0],
                      **kwargs):
        
        if evaluationName not in self.setup:
            self.setup[evaluationName] = PressureField(Length, Width, step, 
                                                        plane, offset)
        elif evaluationName in self.setup and "overwrite" in kwargs:
            self.setup.pop(evaluationName)
            self.setup[evaluationName] = PressureField(Length, Width, step, 
                                                        plane, offset)
        else:
             print("evaluation {} already exists. You can overwrite it using \
                   the 'overwrite' flag".format(evaluationName))

        
    def fieldPoint(self, evaluationName, microphonePosition, **kwargs):        
        if evaluationName not in self.setup:
            self.setup[evaluationName] = FieldPoint(microphonePosition, **kwargs)

        elif evaluationName in self.setup and "overwrite" in kwargs:
            self.setup.pop(evaluationName)
            self.setup[evaluationName] = FieldPoint(microphonePosition, **kwargs)

        else:
             print("evaluation {} already exists. You can overwrite it using \
                   the 'overwrite' flag".format(evaluationName))   


    def boundingBox(self, evaluationName, Lx, Ly, Lz,
                    step=1, offset=[0, 0, 0], **kwargs):
        if evaluationName not in self.setup:
            self.setup[evaluationName] = BoundingBox(Lx, Ly, Lz, step, offset)

        elif evaluationName in self.setup and "overwrite" in kwargs:
            self.setup.pop(evaluationName)
            self.setup[evaluationName] = BoundingBox(Lx, Ly, Lz, step, offset)

        else:
             print("evaluation {} already exists. You can overwrite it using \
                   the 'overwrite' flag".format(evaluationName))   
        
        
    def sphericalRadiation(self, evaluationName, nMic, 
                           radius, offset=[0, 0, 0], **kwargs):
        if evaluationName not in self.setup:
            self.setup[evaluationName] = SphericalRadiation(nMic, radius, offset)

        elif evaluationName in self.setup and "overwrite" in kwargs:
            self.setup.pop(evaluationName)
            self.setup[evaluationName] = SphericalRadiation(nMic, radius, offset)

        else:
             print("evaluation {} already exists. You can overwrite it using \
                   the 'overwrite' flag".format(evaluationName))  
                   
    def plottingGrid(self, evaluationName, path_to_grid, **kwargs):
        if evaluationName not in self.setup:
            self.setup[evaluationName] = PlottingGrid(path_to_grid)

        elif evaluationName in self.setup and "overwrite" in kwargs:
            self.setup.pop(evaluationName)
            self.setup[evaluationName] = PlottingGrid(path_to_grid)

        else:
             print("evaluation {} already exists. You can overwrite it using \
                   the 'overwrite' flag".format(evaluationName))  
    
    def solve(self, evaluation_name="all"):
        if evaluation_name == "all":
            obs_to_compute = []
            for key in self.setup:
                if self.setup[key].isComputed is False:
                    obs_to_compute.append(key)
                else:
                    pass
        elif isinstance(evaluation_name, str):
            obs_to_compute = [evaluation_name]
        elif isinstance(evaluation_name, list):
            obs_to_compute = evaluation_name
        else:
            raise ValueError("evaluation_name should be a key in setup.")
        
        # group microphones to compute in one step (faster for large numbers
        # of microphones)
        mic2compute = np.empty([0, 3])
        for obs in obs_to_compute:
            setup = self.setup[obs]
            xMic = setup.xMic
            nMic = setup.nMic
            if setup.isComputed is False:
                mic2compute = np.concatenate((mic2compute, xMic))
        
        if mic2compute.any() == np.True_:
            _, pMic = self.bemObject.getMicPressure(mic2compute, 
                                                    individualSpeakers=True)
        else:
            None
        
        # ungroup microphones and store within each setup
        current_index = 0
        for obs in obs_to_compute:
            setup = self.setup[obs]
            nMic = setup.nMic
            setup.pMic = pMic[:, current_index:current_index+nMic, :]
            setup.isComputed = True
            current_index += nMic

    def plot_system(self, backend="pyvista"):
        """
        Plot set of evaluations.

        Parameters
        ----------
        backend : str, optional
            Which software is used for display. Can be either "pyvista" or "gmsh".
            The default is "pyvista".

        Returns
        -------
        obj : None
            Plot.

        """
        if backend == "pyvista":
            obj = plot_system_pv(self)
        elif backend == "gmsh":
            obj = plot_system_gmsh(self)
        return obj
    
    def plot(self, evaluations=[], radiatingElement=[], processing=None,
             pf2grid=False, **kwargs):
        """
        Plot evaluations for given radiatingElement

        Parameters
        ----------
        evaluations : str or list of str, optional
            List of evaluations to plot. The default is "all".
        radiatingElement : int or list of int, optional
            List of radiating element to plot. The default is "all"
        processing : postProcessing object
            Post-processing to apply on given surfaces.
        pf2grid : bool
            If True, pressure-fields are "converted" into grids to plot in gmsh.
        **kwargs : key arguments for post-processing options
            for now there is not much options. Mostly for transformation of 
            plotting grids.
                - transformation : str
                    Either 'real', 'imag', 'spl' or 'SPL', 'phase'

        Returns
        -------
        None.

        """
        # Check input for which data to plot
        if bool(evaluations) is False:
            # plot all evaluations
            obs2plot = list(self.setup.keys())
        elif isinstance(evaluations, list):
            obs2plot = evaluations
        else:
            obs2plot = [evaluations]
        
        if bool(radiatingElement) is False:
            # plot all elements
            element2plot = self.bemObject.radiatingElement
        else:
            element2plot = radiatingElement
        
        if processing is not None:
            elementCoeff = np.ones([len(self.frequency), len(element2plot)], 
                                   dtype=complex)
            pp = processing
            for name in pp.TF:
                for idx, element in enumerate(element2plot):
                    if element in pp.TF[name]["radiatingElement"]:
                        elementCoeff[:, idx] *= pp.TF[name]["H"]
                        
        else:
            elementCoeff = np.ones([len(self.frequency), len(element2plot)], 
                                   dtype=complex)
        
        # Sort evaluations by type -> this could def be better, but it works sooo...
        polar, polarName = [], []
        field, pmicField, L, W, xMic_f = [], [], [], [], []
        point = []
        box = []
        sphere = []
        grid = []
        gridName = []
        for key in obs2plot:
            setup = self.setup[key]
            if setup.type == "polar":
                polar.append(setup)
                polarName.append(key)
            elif setup.type == "pressure_field" and pf2grid is False:
                field.append(setup)
                pmicField.append(setup.pMic)
                xMic_f.append(setup.xMic)
                L.append(setup.L)
                W.append(setup.W)
            elif setup.type == "box":
                box.append(setup)
            elif setup.type == "spherical":
                sphere.append(setup)
            elif setup.type == "field_point":
                point.append(setup)
            elif setup.type == "grid":
                grid.append(setup)
                gridName.append(key)
            elif setup.type == "pressure_field" and pf2grid is True:
                grid.append(setup)
                gridName.append(key)
        
        # plot polar
        for i, obs in enumerate(polar):
            pmic2plot = getPressure(obs.pMic, self.bemObject.radiatingElement, 
                                    element2plot, elementCoeff)
            gplot.directivityViewer(obs.theta, self.frequency, pmic2plot,
                                    title=polarName[i])
            
        # plot field
        if bool(field) is True:
            field2plot = getPressure(pmicField, self.bemObject.radiatingElement, 
                                    element2plot, elementCoeff)
            gplot.pressureField_3D(self.bemObject, 
                                   xMic_f, L, W, field2plot, element2plot)
        
        # plot evaluation points
        for i, obs in enumerate(point):
            pMic = obs.pMic
            point2plot = getPressure(pMic, self.bemObject.radiatingElement, 
                                    element2plot, elementCoeff)
            gplot.FRF(self.frequency, point2plot, legend=obs.legend)
            
        # plot spherical radiation
        for i, obs in enumerate(sphere):
            pMic = obs.pMic
            xMic = obs.xMic
            point2plot = getPressure(pMic, self.bemObject.radiatingElement,
                                     element2plot, elementCoeff)
            gplot.sphericalRadiation(xMic, point2plot, self.frequency)
            
        # plot bounding box
        for i, obs in enumerate(box):
            nx, ny, nz = obs.nx, obs.ny, obs.nz
            pMic = obs.pMic
            xMic = obs.xMic
            point2plot = getPressure(pMic, self.bemObject.radiatingElement,
                                     element2plot, elementCoeff)
            gplot.boundingBox(self.bemObject, nx, ny, nz, point2plot,
                              xMic, radiatingElement)
            
        if bool(grid) is True:
            point2plot = []

            for i, obs in enumerate(grid):
                pMic = obs.pMic
                point2plot.append(getPressure(pMic, self.bemObject.radiatingElement,
                                              element2plot, elementCoeff))
            
            if "transformation" in kwargs:
                transformation = kwargs["transformation"]
            else:
                transformation = "SPL"
            if "export_grid" in kwargs:
                export_grid = kwargs["export_grid"]
            else:
                export_grid = False
            gplot.bempp_grid(self.bemObject, grid, point2plot,
                             elementCoeff, element2plot, gridName, 
                             transformation, export_grid)
        return None
    
    


#%% helper function
def getPressure(pmic, radiatingElement, element2plot, coefficients):
    """
    get the pressure to plot given element2plot and their corresponding 
    coefficients

    Parameters
    ----------
    pmic : numpy array
        pressure returned by one evaluation.
    radiatingElement : list of int
        radiatingElement of the BEM object. Used to associate pMic to the 
        correct coefficients
    element2plot : list of int
        specific radiating element to plot.
    coefficients : numpy array
        coefficients to apply to corresponding radiating surfaces.

    Returns
    -------
    pmic_out : TYPE
        DESCRIPTION.

    """
    if isinstance(pmic, list):
        pmic_out = []
        for p in range(len(pmic)):
            nFreq, nMic, nRad = pmic[p].shape
            pmic_tmp = np.zeros([nFreq, nMic], dtype=complex)
            for i, e in enumerate(element2plot):
                for mic in range(nMic):
                    radE = np.argwhere(e == np.array(radiatingElement))[0][0]
                    pmic_tmp[:, mic] += pmic[p][:, mic, radE] * coefficients[:, i]
            pmic_out.append(pmic_tmp)
    else:      
        nFreq, nMic, nRad = pmic.shape
        pmic_out = np.zeros([nFreq, nMic], dtype=complex)
        for i, e in enumerate(element2plot):
            for mic in range(nMic):
                radE = np.argwhere(e == np.array(radiatingElement))[0][0]
                pmic_out[:, mic] += pmic[:, mic, radE] * coefficients[:, i]
    return pmic_out
            

#%% Evaluation classes
class PolarRadiation:
    def __init__(self, minAngle: float,
                 maxAngle: float,
                 step: float,
                 on_axis: str,
                 direction: str,
                 radius: float = 5,
                 offset: list = [0, 0, 0]):
        self.minAngle = minAngle
        self.maxAngle = maxAngle
        self.step = step
        self.on_axis = on_axis
        self.direction = direction
        self.radius = radius
        self.offset = offset
        
        from electroacPy.general.geometry import create_circular_array
        self.theta = np.arange(minAngle, maxAngle+step, step)
        self.xMic = create_circular_array(self.theta, 
                                          on_axis, direction, radius, offset)
        self.nMic = len(self.xMic)
        
        # edit evaluationParameters
        self.type = "polar"
        self.isComputed = False
        self.pMic = None
        
        

class FieldPoint:
    def __init__(self, micPositions, **kwargs): 
        if isinstance(micPositions, np.ndarray):
            self.xMic = micPositions    
        else:
            self.xMic = np.array(micPositions)
        self.nMic = len(self.xMic)
        self.legend = None
        
        if "legend" in kwargs:
            self.legend = kwargs["legend"]
        
        # edit evaluationParameters
        self.type = "field_point"
        self.isComputed = False
        self.pMic = None
        
        
class PressureField:
    def __init__(self, Length: float, 
                 Width: float,
                 step: float,
                 plane: str,
                 offset: list = [0, 0, 0]):
        self.Width = Width
        self.Length = Length
        self.step = step
        self.plane = plane
        self.offset = offset
        
        from electroacPy.general.geometry import create_planar_array
        self.xMic, self.L, self.W = create_planar_array(Length, Width, 
                                                        step, plane, offset)        
        
        self.nMic = len(self.xMic)
            
        # help with dimensions - might need to change
        self.nx, self.ny, self.nz = 0, 0, 0
        L, W = len(self.L), len(self.W)
        
        # Mapping planes to dimensions
        plane_mapping = {
            ('x', 'y'): (L, W, 1),
            ('x', 'z'): (L, 1, W),
            ('y', 'x'): (W, L, 1),
            ('y', 'z'): (1, L, W),
            ('z', 'x'): (W, 1, L),
            ('z', 'y'): (1, W, L)
        }
        
        # Set dimensions based on the selected plane
        self.nx, self.ny, self.nz = plane_mapping.get((self.plane[0], 
                                                      self.plane[1]), 
                                                      (0, 0, 0))

        # edit evaluationParameters
        self.type = "pressure_field"
        self.isComputed = False
        self.pMic = None
        
        # data for gmsh plots
        from scipy.spatial import Delaunay
        if plane == "xy" or plane == "yx":
            dim_ = [0, 1]
        elif plane == "xz" or plane == "zx":
            dim_ = [0, 2]
        elif plane == "yz" or plane == "zy":
            dim_ = [1, 2]
    
        triangulation = Delaunay(self.xMic[:, dim_])
        self.elements = triangulation.simplices.T
        self.dof      = self.nMic
        self.vertices = self.xMic.T

class BoundingBox:
    def __init__(self, Lx, Ly, Lz, step=1, offset=[0, 0, 0]):
        self.Lx = Lx
        self.Ly = Ly 
        self.Lz = Lz
        self.step = step
        self.offset = offset
        
        from electroacPy.general.geometry import create_bounding_box
        self.xMic, dim = create_bounding_box(Lx, Ly, Lz, step, offset)
        self.nMic = len(self.xMic)
        self.nx = dim[0]
        self.ny = dim[1]
        self.nz = dim[2]

        # edit evaluationParameters
        self.type = "box"
        self.isComputed = False
        self.pMic = None
        
        
class SphericalRadiation:
    def __init__(self, nMic, radius, offset):
        self.radius = radius
        self.offset = offset
        
        from electroacPy.general.geometry import create_spherical_array
        self.xMic = create_spherical_array(nMic, radius, offset)
        self.nMic = len(self.xMic)

        # edit evaluationParameters
        self.type = "spherical"
        self.isComputed = False
        self.pMic = None
    

class PlottingGrid:
    def __init__(self, path_to_grid):
        from bempp.api import import_grid
        from ..general.geometry import check_mesh
        
        self.path_to_grid = check_mesh(path_to_grid)
        grid = import_grid(self.path_to_grid)
        self.dof      = grid.vertices.shape[1]
        self.vertices = grid.vertices
        self.elements = grid.elements
        self.edges    = grid.edges[:, grid.edge_on_boundary]
        # self.edges_on_boundary = grid.edge_on_boundary
        self.xMic     = grid.vertices.T
        self.nMic     = len(self.xMic)
        
        # edit evaluationParameters
        self.type = "grid"
        self.isComputed = False
        self.pMic = None
        
#%% Plotting help

def create_boundary_planes(boundary, offset, plane_size=10):
    """
    Create PyVista PolyData for boundaries with adjusted positioning to form corners.

    Parameters:
        boundary (list): List of boundary normals (e.g., ["x", "y", "z"]).
        offset (list): List of offsets corresponding to each boundary normal.
        plane_size (float): Maximum size of the plane to be created (e.g., 10m).

    Returns:
        pyvista.PolyData: PolyData object representing the boundary planes.
    """
    
    if len(boundary) != len(offset):
        raise ValueError("The boundary and offset lists must have the same length.")
    
    # Initialize an empty PolyData object to collect all the planes
    planes = pyvista.PolyData()
    
    # Define default plane extent (10m by 10m)    

    # Loop through each boundary and create a corresponding plane with adjusted size
    for i, axis in enumerate(boundary):
        # Create the points for the plane depending on the axis
        if axis in ["x", "X"]:
            # Plane orthogonal to the x-axis (yz-plane)
            x_center = offset[i]  # Set x at the offset
            
            if "y" in boundary:
                y_offset = offset[boundary.index("y")]
                y_center = (plane_size[1]-y_offset)/2 + y_offset
            elif "Y" in boundary:
                y_offset = offset[boundary.index("Y")]
                y_center = (plane_size[1]-y_offset)/2 + y_offset
            else:
                y_offset=0
                y_center=0
            
            if "z" in boundary:
                z_offset = offset[boundary.index("z")]
                z_center = (plane_size[2]-z_offset)/2 + z_offset
            elif "Z" in boundary:
                z_offset = offset[boundary.index("Z")]
                z_center = (plane_size[2]-z_offset)/2 + z_offset
            else:
                z_offset=0
                z_center=0

            point = [x_center, y_center, z_center]  # Plane's center
            normal = [1, 0, 0]  # Normal vector in x-direction
            
            plane = pyvista.Plane(center=point, direction=normal, 
                             i_size=plane_size[2]-z_offset, 
                             j_size=plane_size[1]-y_offset)
        
        elif axis in ["y", "Y"]:
            # Plane orthogonal to the y-axis (xz-plane)
            y_center = offset[i]  # Set y at the offset
            
            if "x" in boundary:
                x_offset = offset[boundary.index("x")]
                x_center = (plane_size[0]-x_offset)/2 + x_offset
            elif "X" in boundary:
                x_offset = offset[boundary.index("X")]
                x_center = (plane_size[0]-x_offset)/2 + x_offset
            else:
                x_offset=0
                x_center=0
            
            if "z" in boundary:
                z_offset = offset[boundary.index("z")]
                z_center = (plane_size[2]-z_offset)/2 + z_offset
            elif "Z" in boundary:
                z_offset = offset[boundary.index("Z")]
                z_center = (plane_size[2]-z_offset)/2 + z_offset
            else:
                z_offset=0
                z_center=0
            
            point = [x_center, y_center, z_center]  # Plane's center
            normal = [0, 1, 0]  # Normal vector in x-direction
            
            plane = pyvista.Plane(center=point, direction=normal, 
                             i_size=plane_size[2]-z_offset, 
                             j_size=plane_size[0]-x_offset)
        
        elif axis in ["z", "Z"]:
            # Plane orthogonal to the z-axis (xy-plane)
            z_center = offset[i]  # Set z at the offset
            
            if "x" in boundary:
                x_offset = offset[boundary.index("x")]
                x_center = (plane_size[0]-x_offset)/2 + x_offset
            elif "X" in boundary:
                x_offset = offset[boundary.index("X")]
                x_center = (plane_size[0]-x_offset)/2 + x_offset
            else:
                x_offset=0
                x_center=0
            
            if "y" in boundary:
                y_offset = offset[boundary.index("y")]
                y_center = (plane_size[1]-y_offset)/2 + y_offset
            elif "Y" in boundary:
                y_offset = offset[boundary.index("Y")]
                y_center = (plane_size[1]-y_offset)/2 + y_offset
            else:
                y_offset=0
                y_center=0
            
            point = [x_center, y_center, z_center]  # Plane's center
            normal = [0, 0, 1]  # Normal vector in x-direction
            
            plane = pyvista.Plane(center=point, direction=normal, 
                             i_size=plane_size[0]-x_offset,
                             j_size=plane_size[1]-y_offset)
        
        else:
            raise ValueError(f"Unknown boundary axis: {axis}. Must be 'x', 'y', or 'z'.")
        
        # Append the generated plane to the overall PolyData object
        planes += plane
    
    return planes
        
 
#%% PLOT SYSTEM
def plot_system_pv(eval_obj):
    """
    Plot current evaluation points in PyVista.
    
    Parameters
    ----------
    eval_obj: evaluation object
        Display set of evaluation.
    """
    
    if hasattr(eval_obj.bemObject, "xSource") and eval_obj.bemObject.meshPath is None:
        pl = pyvista.Plotter()
        pl.disable_anti_aliasing()
        xSource = eval_obj.bemObject.xSource
        point_cloud_source = pyvista.PolyData(xSource.astype(float))
        pl.add_mesh(point_cloud_source, color="black",
                    render_points_as_spheres=True, label="Sources", point_size=6)
        boundary_parameters = eval_obj.bemObject.boundary_conditions.parameters
    else:
        mesh = pyvista.read(eval_obj.bemObject.meshPath)
        radSurf = eval_obj.bemObject.radiatingElement
        boundary_parameters = eval_obj.bemObject.boundary_conditions
        
        # prepare mesh color, camera and bounds
        colors = []
        for scalar in mesh.active_scalars:
            if scalar in radSurf:
                colors.append([255, 0, 0])  # Red color for radiators
            else:
                colors.append([128, 128, 128])  # gray color for non-radiating
        mesh.cell_data['colors'] = colors
    
        # create plotter
        pl = pyvista.Plotter()
        pl.disable_anti_aliasing()
        if eval_obj.bemObject.domain == "exterior":
            pl.add_mesh(mesh, show_edges=True, cmap='summer', scalars='colors',
                        show_scalar_bar=False, rgb=True)
        elif eval_obj.bemObject.domain == "interior":
            pl.add_mesh_clip_box(mesh, show_edges=True, cmap='summer',  
                                 scalars='colors', show_scalar_bar=False, 
                                 rgb=True, rotation_enabled=False)
        light = pyvista.Light(light_type='headlight')
        pl.add_light(light)
    
    
    
    # add evaluation points
    colour = ['blue', 'red', 'green', 'orange', 
              'purple', 'brown', 'yellow', 'cyan'] * 4
    for i, key in enumerate(eval_obj.setup):
        setup = eval_obj.setup[key]
        xMic = setup.xMic
        point_cloud_tmp = pyvista.PolyData(xMic.astype(float))
        if setup.type == "box":
            pl.add_mesh(point_cloud_tmp.outline(), color=colour[i],
                        render_points_as_spheres=True, label=key, point_size=6)
        elif setup.type == "pressure_field":
            grid = pyvista.StructuredGrid()
            grid.points = xMic
            grid.dimensions = [setup.nx, setup.ny, setup.nz]
            pl.add_mesh(grid, show_edges=True, 
                        style='wireframe', color=colour[i], label=key)
        else:
            pl.add_mesh(point_cloud_tmp, color=colour[i],
                        render_points_as_spheres=True, label=key, point_size=6)

    # get bounds to add planes
    bounds = pl.bounds
    x_width = bounds[1] - bounds[0]  # Width along the x-axis
    y_width = bounds[3] - bounds[2]  # Width along the y-axis
    z_width = bounds[5] - bounds[4]  # Width along the z-axis
    
    if bool(eval_obj.setup) is False:   
        p_coeff = 2  # make infinite boundary a bit bigger if no eval is set 
    else:
        p_coeff = 1
    
    plane_size = (x_width*p_coeff, y_width*p_coeff, z_width*p_coeff)
    # add floor if infinite boundary conditions are present in bemObject
    bemObject = eval_obj.bemObject
    if bemObject.boundary_conditions is not None:
        inf_bc = []
        offset = []
        for key in boundary_parameters:
            if key in ["x", "X", "y", "Y", "z", "Z"]:
                inf_bc.append(key)
                offset.append(boundary_parameters[key]["offset"]) 
        
        if bool(inf_bc) is True:
            boundary_planes = create_boundary_planes(inf_bc, offset, plane_size)
            # Add boundary planes to plotter
            pl.add_mesh(boundary_planes, color="gray", opacity=0.5)
        else:
            pass

    if bool(eval_obj.setup) is True:   
        pl.add_legend(face='circle', bcolor=None)

    pl.add_axes(color='black')
    pl.background_color = 'white'

    _ = pl.show_grid(color='k')
    obj = pl.show()
    return obj


def plot_system_gmsh(eval_obj):
    """
    Plot current evaluation points in GMSH.
    
    Parameters
    ----------
    eval_obj: evaluation object
        Display set of evaluation.
    """
    import gmsh
    gmsh.initialize()
    
    if hasattr(eval_obj.bemObject, "xSource") and eval_obj.bemObject.meshPath is None:
        source_group = []
        for i in range(len(eval_obj.bemObject.xSource)):
            xSource = eval_obj.bemObject.xSource[i, :]
            point = gmsh.model.geo.add_point(xSource[0], xSource[1], xSource[2])
            source_group.append(point) 
            gmsh.model.geo.synchronize()
            gmsh.model.setColor([(0, point)], 0, 0, 0)
        physical_group_id = gmsh.model.addPhysicalGroup(0, source_group)  # '0' stands for point entities
        gmsh.model.setPhysicalName(0, physical_group_id, "sources")  # Optional: name the physical group
    elif hasattr(eval_obj.bemObject, "xSource") and eval_obj.bemObject.meshPath is not None:
        gmsh.open(eval_obj.bemObject.meshPath)
        source_group = []
        for i in range(len(eval_obj.bemObject.xSource)):
            xSource = eval_obj.bemObject.xSource[i, :]
            point = gmsh.model.geo.add_point(xSource[0], xSource[1], xSource[2])
            source_group.append(point) 
            gmsh.model.geo.synchronize()
            gmsh.model.setColor([(0, point)], 0, 0, 0)
        physical_group_id = gmsh.model.addPhysicalGroup(0, source_group)  # '0' stands for point entities
        gmsh.model.setPhysicalName(0, physical_group_id, "sources")  # Optional: name the physical group
    else:
        gmsh.open(eval_obj.bemObject.meshPath)
  

    # get max nodes in system
    all_nodes = gmsh.model.mesh.getNodes()
    maxNode =  []
    for i in all_nodes:
        try:
            maxNode.append(np.max(i))
        except:
            None
    maxNODE = int(np.max(maxNode)) + 1
    
    colors = [
        [0, 0, 255],    # Blue
        [255, 165, 0],  # Orange
        [0, 128, 0],    # Green
        [255, 0, 0],    # Red
        [128, 0, 128],  # Purple
        [165, 42, 42],  # Brown
        [255, 192, 203],# Pink
        [128, 128, 128],# Grey
        [128, 128, 0],  # Olive
        [0, 255, 255],  # Cyan
    ]

    # Function to repeat the colors to match the number of observations
    def get_colors(num_observations):
        return [colors[i % len(colors)] for i in range(num_observations)]

    nEval = len(eval_obj.setup)
    colorsEval = get_colors(nEval) 

    for s, setup in enumerate(eval_obj.setup):
        cl = colorsEval[s]
        csetup = eval_obj.setup[setup]
        obs_group = []
        nMic = csetup.nMic
        xMic = csetup.xMic
        if csetup.type not in ["grid", "pressure_field"]:
            for i in range(nMic):
                point = gmsh.model.geo.add_point(xMic[i, 0], xMic[i, 1], xMic[i, 2])
                obs_group.append(point) 
                gmsh.model.geo.synchronize()
                gmsh.model.setColor([(0, point)], cl[0], cl[1], cl[2])     # Red for point p1
            physical_group_id = gmsh.model.addPhysicalGroup(0, obs_group)  # '0' stands for point entities
            gmsh.model.setPhysicalName(0, physical_group_id, setup)        # Optional: name the physical group
        
        elif csetup.type in ["grid", "pressure_field"]:
            # node
            nodeTags_n = []
            coord = []
            
            try: # in case plottingGrids are set before other evaluations
                maxDim0 = int(np.max(gmsh.model.getEntities(0)) + 1)
            except:
                maxDim0 = 1    
            maxDim2 = int(np.max(gmsh.model.getEntities(2)) + 1)
            
        
            # elements
            elements              = csetup.elements
            nElements             = elements.shape[1]
            nodeTags_elements     = []
            elementTags_elements  = []
            
            for i in range(nMic):
                nodeTags_n.append(i+maxNODE)
                coord.append(xMic[i, 0])
                coord.append(xMic[i, 1])
                coord.append(xMic[i, 2])
        
            # elements of dim 2
            for k in range(nElements):
                elementTags_elements.append(k+1)
                nodeTags_elements.append(elements[0, k]+maxNODE) 
                nodeTags_elements.append(elements[1, k]+maxNODE) 
                nodeTags_elements.append(elements[2, k]+maxNODE)
                

            gmsh.model.addDiscreteEntity(0, maxDim0)
            gmsh.model.addDiscreteEntity(2, maxDim2)
            gmsh.model.mesh.addNodes(0, maxDim0, nodeTags_n, coord)
            gmsh.model.mesh.addElements(2, maxDim2, [2], 
                                                    [elementTags_elements], [nodeTags_elements])
            
            physical_group_id = gmsh.model.addPhysicalGroup(2, [maxDim2])  # '0' stands for point entities
            gmsh.model.setPhysicalName(2, physical_group_id, setup) 
            
            maxNODE += i
    
    gmsh.option.setNumber("Geometry.PointSize", 3)
    gmsh.option.setNumber("Geometry.PointType", 1)
    gmsh.model.geo.synchronize()

    gmsh.fltk.run()

    gmsh.finalize()
    return None



def create_gmsh_mesh(xMic, elements, mesh_filename="mesh.msh"):
    import gmsh
    gmsh.initialize()
    gmsh.model.add("mic_mesh")

    # Add nodes
    nMic = xMic.shape[0]
    for i in range(nMic):
        gmsh.model.mesh.addNode(i + 1, xMic[i, 0], xMic[i, 1], xMic[i, 2])

    # Add triangular elements
    nElements = elements.shape[1]
    elementType = gmsh.model.mesh.getElementType("triangle", 1)  # 2D triangle element
    gmsh.model.mesh.addElements(2, [elementType], [[i + 1 for i in range(nElements)]], elements + 1)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  3 15:50:43 2023

@author: tom.munoz
"""
import numpy as np
import electroacPy.general as gtb
from electroacPy.global_ import air
from electroacPy.general import lp_loaders as lpl
import matplotlib.pyplot as plt
import tkinter as tk
from tkinter import ttk
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure
import os
from copy import copy

pi = np.pi

class electroAcousticDriver:
    def __init__(self, U, Le, Re, Cms, Mms, Rms, Bl, Sd, f_array,
                 c=air.c, rho=air.rho):
        """
        Create an electro-acoustic driver from its Thiele-Small parameters.

        Parameters
        ----------
        U : float
            Input voltage (usually 1).
        Le : float
            Coil inductance (H).
        Re : float
            Coil resistance (Ohms).
        Cms : float
            Suspension's compliance (m/N).
        Mms : float
            Moving mass of the driver (kg).
        Rms : float
            Mechanical resistance (N.s/m).
        Bl : float
            Force factor (N.A / T.m).
        Sd : float
            Diaphragm surface area (m^2).
        f_array : array-like
            Frequencies at which the solutions are computed.
        c : float, optional
            Speed of sound. The default is air.c.
        rho : float, optional
            Density of air. The default is air.rho.

        Returns
        -------
        None

        Notes
        -----
        This class represents an electro-acoustic driver and calculates its electrical and mechanical impedance,
        equivalent acoustical impedance, input pressure, radiation impedance, quality factors, and other parameters.

        """
        # identifier
        self.identifier = "EAC"

        # medium properties
        w = 2*pi*f_array
        Zc = rho * c
        k = w / c
        
        # self param
        self.U = U
        self.Le = Le
        self.Re = Re
        self.Cms = Cms
        self.Mms = Mms
        self.Rms = Rms
        self.Bl = Bl
        self.Sd = Sd
        self.f_array = f_array

        # medium returns
        self.c = c
        self.k = k
        self.f = f_array
        self.w = w
        self.rho = rho

        # speaker radius
        r = np.sqrt(Sd / pi)
        self.r = r
        self.Sd = Sd

        # impedance
        s = 1j*w
        self.Ze = Re + s*Le
        Zms = Rms + s*Mms + 1/(s*Cms)

        # equivalent acoustical impedance
        self.Zac = (1/Sd**2)*(Bl**2/self.Ze)
        self.Zas = Zms / Sd**2
        self.Zms = Zms
        self.ZeTot = self.Ze + Bl**2/Zms
        self.Bl = Bl
        # equivalent input pressure
        self.Ps = U * Bl / (self.Ze * Sd)

        # Radiation impedance (speaker front impedance)
        Mrad = 8 * rho * r / 3 / pi / Sd
        Rrad = Zc / Sd * (k*r)**2 / 2
        self.Zaf = Rrad + 1j*w*Mrad
        self.Zs = self.Zac + self.Zas + self.Zaf # total acoustical impedance coming from the driver

        # Quality factors and others
        self.Qes = Re / (Bl)**2 * np.sqrt(Mms/Cms)
        self.Qms = 1/Rms * np.sqrt(Mms/Cms)
        self.Qts = self.Qms*self.Qes / (self.Qms + self.Qes)
        self.Vas = rho*c**2*Sd**2*Cms
        self.Fs = 1 / (2*pi*np.sqrt(Cms*Mms))
        self.EBP = self.Fs/self.Qes


        # Ref Signals
        # Velocity
        self.Hv = Bl/self.Ze  / (self.Zms + Bl**2 / self.Ze) * self.U

        # Displacement
        self.Hx = self.Hv / s

        # Acceleration
        self.Ha = self.Hv * s

        # Acoustic simulation reference
        self.ref2bem = None

        # in box velocity and impedance
        self.inBox    = False
        self.isPorted = False  # easier to manage if radiator in study_ is not a speakerBox object
        self.v        = self.Hv 
        self.Q        = self.v * self.Sd

        # references
        self.ref2bem   = False
        self.poly_data = False  # is class from polytech?

        
    def plotZe(self, **kwargs):
        """
        Plot the electrical impedance ZeTot in both modulus and phase.

        Returns
        -------
        None

        """
        
        if "figsize" in kwargs:
            size=kwargs["figsize"]
        else:
            size=None
        
        fig, ax = plt.subplots(2, 1, figsize=size)
        ax[0].semilogx(self.f_array, np.abs(self.ZeTot))
        ax[0].set(ylabel="Magnitude [Ohm]")
        
        ax[1].semilogx(self.f_array, np.angle(self.ZeTot))
        ax[1].set(xlabel="Frequency [Hz]", ylabel="Phase [rad]")
        for i in range(2):
            ax[i].grid(which="both", linestyle="dotted")
        plt.tight_layout()
        
        if "savefig" in kwargs:
            path = kwargs["savefig"]
            plt.savefig(path)
        return plt.show()
    
    
    def plotXVA(self, **kwargs):
        """
        Plot the displacement, velocity, and acceleration frequency responses.

        Returns
        -------
        None

        """
        
        if "figsize" in kwargs:
            size=kwargs["figsize"]
        else:
            size=None
        
        fig, ax = plt.subplots(3, 1, figsize=size)
        ax[0].semilogx(self.f_array, np.abs(self.Hx*1e3), label='Displacement')
        ax[1].semilogx(self.f_array, np.abs(self.Hv), label='Velocity')
        ax[2].semilogx(self.f_array, np.abs(self.Ha), label='Acceleration')
        ax[2].set(xlabel="Frequency [Hz]")
        ax[0].set(ylabel="mm")
        ax[1].set(ylabel="m/s")
        ax[2].set(ylabel="m/s^2")
        for i in range(3):
            ax[i].grid(which='both', linestyle="dotted")
            ax[i].legend(loc='best')
        plt.tight_layout()
       
        if "savefig" in kwargs:
            path = kwargs["savefig"]
            plt.savefig(path)
      
        return plt.show()

    def getThieleSmallParam(self):
        """
        Print out the Thiele/Small parameters of the electro-acoustic driver.

        Returns
        -------
        None

        """
        greetingStr = "Thiele/Small parameters"
        print(greetingStr)
        print("-"*len(greetingStr))
        print("--- Electrical ---")
        print("Re = ", self.Re, " Ohm")
        print("Le = ", self.Le*1e3, " mH")
        print("Bl = ", self.Bl, " N/A")
        
        print("--- Mechanical ---")
        print("Rms = ", round(self.Rms, 2), " N.s/m")
        print("Mms = ", round(self.Mms*1e3, 2), " g")
        print("Cms = ", round(self.Cms*1e3, 2), " mm/N")
        print("Kms = ", round(1/self.Cms), "N/m")
        print("Sd = ", round(self.Sd*1e4, 2), " cm^2")
    
        print("--- Quality Factors ---")
        print("Qes = ", round(self.Qes, 2))
        print("Qms = ", round(self.Qms, 2))
        print("Qts = ", round(self.Qts, 2))
        
        print("--- Others ---")
        print("Fs = ", round(self.Fs, 2), " Hz")
        print("Vas = ", self.Vas, " m^3")
        return None
    
    def sealedAlignment(self):
        """
        Compute Volume from Qtc value using Tkinter instead of Matplotlib widgets
    
        Parameters
        ----------
        driver : class
            electro_acoustic_driver object.
        Qtc : total quality factor (mechanical, electrical, acoustical)
        c : speed of sound. The default is air.c.
        rho : air density. The default is air.rho.
    
        Returns
        -------
        Vb : sealed enclosure volume.
        fc : resonance frequency of the driver inside the enclosure (without radiation mass)
        """
        
        driver = self
        c = self.c
        rho = self.rho
    
        ## box parameters
        Vb = driver.Vas
        fc = driver.Fs * np.sqrt(driver.Vas / Vb + 1)
        Qtc = fc / driver.Fs * driver.Qts
        Cab = Vb / rho / c**2
    
        ## radiated pressure at 1 m
        f_axis = driver.f_array
        omega = 2 * np.pi * f_axis
        k = omega / c
    
        Zac = driver.Zac
        Zas = driver.Zas
        Zab = 1 / 1j / omega / Cab
        Ps = driver.Ps
        Qs = Ps / (Zac + Zas + Zab)  # removed Zaf
    
        p = 1j * k * rho * c * Qs * np.exp(-1j * k * 1) / (2 * np.pi * 1)
        Ze = driver.Ze + driver.Bl ** 2 / (driver.Zms + driver.Sd ** 2 * (Zab))
    
        # Setup Tkinter window
        root = tk.Tk()
        root.title("Sealed Alignment")
    
        # Create figure (using matplotlib's Figure class, not plt.subplots)
        fig = Figure(figsize=(6, 4))
        ax1 = fig.add_subplot(211)
        ax2 = fig.add_subplot(212)
        
        # Initial plot (SPL and Magnitude)
        ax1.semilogx(f_axis, gtb.gain.SPL(p), label='SPL')
        ax2.semilogx(f_axis, np.abs(Ze), label='Magnitude')
    
        ax2.set(xlabel='Frequency [Hz]', ylabel='Impedance')
        ax1.set(ylabel='SPL [dB] at 1 meter')
        ax1.legend(loc='best')
        ax2.legend(loc='best')
    
        for ax in [ax1, ax2]:
            ax.grid(which='both')
    
        # Embed the plot into Tkinter
        canvas = FigureCanvasTkAgg(fig, master=root)
        canvas.draw()
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)
    
        # Create Labels and Entry fields
        defaultQtc = round(Qtc, 4)
    
        label_qtc = ttk.Label(root, text="Qtc:")
        label_qtc.pack(side=tk.LEFT, padx=5)
        
        entry_qtc = ttk.Entry(root, width=10)
        entry_qtc.pack(side=tk.LEFT, padx=5)
        entry_qtc.insert(0, str(defaultQtc))
    
        label_vb = ttk.Label(root, text="Vb (L):")
        label_vb.pack(side=tk.LEFT, padx=5)
        
        entry_vb = ttk.Entry(root, width=10)
        entry_vb.pack(side=tk.LEFT, padx=5)
        entry_vb.insert(0, str(round(Vb * 1e3, 4)))
    
        # Function to update the plot based on Qtc entry
        def update_plot():
            try:
                Qtc_value = float(entry_qtc.get())
                fc = Qtc_value / driver.Qts * driver.Fs
                Vb_new = driver.Vas / ((fc / driver.Fs)**2 - 1)
                Cab_new = Vb_new / rho / c**2
                Zab_new = 1 / 1j / omega / Cab_new
    
                Qs_new = Ps / (Zac + Zas + Zab_new)
                p_new = 1j * k * rho * c * Qs_new * np.exp(-1j * k * 1) / (2 * np.pi * 1)
                Ze_new = driver.Ze + driver.Bl ** 2 / (driver.Zms + driver.Sd ** 2 * Zab_new)
    
                # Clear and update the plots
                ax1.clear()
                ax2.clear()
    
                ax1.semilogx(f_axis, gtb.gain.SPL(p_new), label='SPL')
                ax2.semilogx(f_axis, np.abs(Ze_new), label='Magnitude')
    
                ax2.set(xlabel='Frequency [Hz]', ylabel='Impedance [Ohm]')
                ax1.set(ylabel='SPL [dB] at 1 meter')
                ax1.legend(loc='best')
                ax2.legend(loc='upper left')
    
                for ax in [ax1, ax2]:
                    ax.grid(which='both')
    
                canvas.draw()
    
                # Update volume value in the entry box
                entry_vb.delete(0, tk.END)
                entry_vb.insert(0, str(round(Vb_new * 1e3, 4)))
    
            except ValueError:
                pass  # Prevent the function from crashing if non-numeric input is entered
    
        # Function to update Qtc based on Volume entry
        def update_qtc():
            try:
                Vb_value = float(entry_vb.get()) * 1e-3  # Convert from liters to cubic meters
                fc_new = driver.Fs * np.sqrt(driver.Vas / Vb_value + 1)
                Qtc_new = fc_new / driver.Fs * driver.Qts
    
                # Update Qtc value in the entry box
                entry_qtc.delete(0, tk.END)
                entry_qtc.insert(0, str(round(Qtc_new, 4)))
    
                update_plot()  # Automatically update plot with new values
    
            except ValueError:
                pass  # Handle non-numeric input
    
        # Bind events to update the plot automatically when the user presses Enter or leaves the entry field
        entry_qtc.bind("<Return>", lambda event: update_plot())
        entry_qtc.bind("<FocusOut>", lambda event: update_plot())
        
        entry_vb.bind("<Return>", lambda event: update_qtc())
        entry_vb.bind("<FocusOut>", lambda event: update_qtc())
    
        root.mainloop()
     
    
    def portedAlignment(self):
        driver = copy(self)
        c = self.c
        rho = self.rho
        f_axis = driver.f_array
        omega = 2 * np.pi * f_axis
        k = omega / c
        s = 1j * omega
        eta = 1e-5
    
        # Default parameters
        self.Vb = copy(driver.Vas)
        self.Lp = 343 / driver.Fs / 100  # Length in meters
        self.rp = self.Lp / 2  # Radius in meters
        self.Sp = np.pi * self.rp ** 2  # Port cross-sectional area
    
        # Create input widgets
        default_volume = str(round(self.Vb * 1e3, 2))
        default_length = str(round(self.Lp * 1e2, 2))
        default_radius = str(round(self.rp * 1e2, 2))
        default_section = str(round(self.Sp * 1e4, 2))
    
        # GUI creation
        root = tk.Tk()
        root.title("Ported Alignment")
    
        # Create a matplotlib figure
        fig = Figure(figsize=(6, 4))
        ax_spl = fig.add_subplot(211)
        ax_imp = fig.add_subplot(212)
        
        
        # Create canvas for the plot and add to the tkinter window
        canvas = FigureCanvasTkAgg(fig, master=root)
        canvas.draw()
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)
    
        # Create textboxes in a horizontal row at the bottom
        volume_label = ttk.Label(root, text="Vb (L):")
        volume_label.pack(side=tk.LEFT, padx=5)
        volume_entry = ttk.Entry(root, width=10)
        volume_entry.pack(side=tk.LEFT, padx=5)
        volume_entry.insert(0, default_volume)
        
        length_label = ttk.Label(root, text="Lp (cm):")
        length_label.pack(side=tk.LEFT, padx=5)
        length_entry = ttk.Entry(root, width=10)
        length_entry.pack(side=tk.LEFT, padx=5)
        length_entry.insert(0, default_length)
        
        radius_label = ttk.Label(root, text="rp (cm):")
        radius_label.pack(side=tk.LEFT, padx=5)
        radius_entry = ttk.Entry(root, width=10)
        radius_entry.pack(side=tk.LEFT, padx=5)
        radius_entry.insert(0, default_radius)
       
        section_label = ttk.Label(root, text=r"Sp (cm²):")
        section_label.pack(side=tk.LEFT, padx=5)
        section_entry = ttk.Entry(root, width=10)
        section_entry.pack(side=tk.LEFT, padx=5)
        section_entry.insert(0, default_section)
        
        def update_plot():
            try:
                # box impedance
                self.Cab = self.Vb / rho / c ** 2
                self.Rab = rho * c / eta / self.Vb
                self.Zbox = gtb.parallel(1 / s / self.Cab, self.Rab)
        
                # port impedance
                self.Map = rho * self.Lp / self.Sp
                self.Mal = 0.85 * 2 * self.rp
                self.Map += self.Mal
                Pp = 2 * np.pi * self.rp
                alpha = np.sqrt(f_axis) * (0.95e-5 + 2.03e-5) * Pp / 2 / self.Sp
                kl = k * (1 + alpha * (1 - 1j))
                d0 = (0.6133 + 0.85) * self.rp
                self.Zrad = rho * c / self.Sp * (1 / 4 * (kl * self.rp) ** 2 + 1j * kl * d0)
                self.Zp = s * self.Map + self.Zrad
                self.Zab = gtb.parallel(1 / s / self.Cab, self.Rab, s * self.Map + self.Zrad)
    
                # Calculate total system
                ZaTot = driver.Zs
                Ps = driver.Ps
                Qs = Ps / (ZaTot + self.Zab)
                Qp = Qs * self.Zbox / (self.Zbox + self.Zp)
    
                p_s = 1j * k * rho * c * Qs * np.exp(-1j * k * 1) / (2 * np.pi * 1)
                p_p = -1j * k * rho * c * Qp * np.exp(-1j * k * 1) / (2 * np.pi * 1)
                Ze = driver.Ze + driver.Bl ** 2 / (driver.Zms + driver.Sd**2 * (self.Zab + driver.Zaf))
        
                # Clear the axes and plot new data
                ax_spl.clear()
                ax_imp.clear()
                ax_spl.semilogx(f_axis, gtb.gain.SPL(p_s), label='driver')
                ax_spl.semilogx(f_axis, gtb.gain.SPL(p_p), label='port')
                ax_spl.semilogx(f_axis, gtb.gain.SPL(p_s+p_p), label='total')
                ax_imp.semilogx(f_axis, np.abs(Ze), label='Impedance')
                ax_spl.set_ylabel('SPL [dB]')
                ax_spl.set_ylim(np.min(gtb.gain.SPL(p_s+p_p)), 
                                np.max(gtb.gain.SPL(p_s+p_p))+6)
                ax_imp.set_ylabel('Impedance [Ohm]')
                ax_imp.set_xlabel('Frequency [Hz]')
                ax_spl.grid(which='both')
                ax_imp.grid(which='both')
                ax_spl.legend(loc='best')
                ax_imp.legend(loc='best')
                canvas.draw()
            except ValueError:
                pass
                
                
        def update_volume():
            try:
                self.Vb = float(volume_entry.get()) * 1e-3  # Convert L to m^3
                update_plot()
            except ValueError:
                print("Invalid input for volume. Please enter a numeric value.")
        
        def update_length():
            try:
                self.Lp = float(length_entry.get()) * 1e-2  # Convert cm to m
                update_plot()
            except ValueError:
                print("Invalid input for length. Please enter a numeric value.")
        
        def update_radius():
            try:
                self.rp = float(radius_entry.get()) * 1e-2  # Convert cm to m
                self.Sp = np.pi * self.rp ** 2
                section_entry.delete(0, "end")
                section_entry.insert(0, str(round(self.Sp * 1e4, 2)))  # Update section box in cm^2
                update_plot()
            except ValueError:
                print("Invalid input for radius. Please enter a numeric value.")
    
        def update_section():
            try:
                self.Sp = float(section_entry.get()) * 1e-4  # Convert cm² to m²
                self.rp = np.sqrt(self.Sp / np.pi)
                radius_entry.delete(0, "end")
                radius_entry.insert(0, str(round(self.rp * 1e2, 2)))  # Update radius box in cm
                update_plot()
            except ValueError:
                print("Invalid input for section. Please enter a numeric value.")
        
    
        # bind Return and FocusOut
        volume_entry.bind("<Return>", lambda event: update_volume())  # Bind Enter key
        volume_entry.bind("<FocusOut>", lambda event: update_volume())
        length_entry.bind("<Return>", lambda event: update_length())  # Bind Enter key
        length_entry.bind("<FocusOut>", lambda event: update_length())
        radius_entry.bind("<Return>", lambda event: update_radius())
        radius_entry.bind("<FocusOut>", lambda event: update_radius())
        section_entry.bind("<Return>", lambda event: update_section())
        section_entry.bind("<FocusOut>", lambda event: update_section())
      
        # Initial plot
        update_plot()
    
        # Run the tkinter loop
        root.mainloop()
    
    def exportZe(self, folder_name, file_name):
        import os
        
        # Create the folder if it doesn't exist
        if not os.path.exists(folder_name):
            os.makedirs(folder_name)
        
        module = np.abs(self.ZeTot)
        phase = np.angle(self.ZeTot, deg=True)
        path = os.path.join(folder_name, file_name)
        np.savetxt(path, np.array([self.f_array, module, phase]).T,
                   fmt="%.3f", 
                   header="Freq[Hz]  Imp[Ohm]  Phase[Deg]",
                   delimiter=',',
                   comments='')

    

def loadLPM(lpmfile, freq_array, U=1, LeZero=False,
            number_of_drivers=1,
            wiring='parallel',
            c=air.c,
            rho=air.rho):
    
    # define loader based on extension
    _, extension = os.path.splitext(lpmfile)
    if extension == ".qsp":
        loader    = lpl.qspeaker_lp_loader
        weight_Le  = 1e-3
        weight_Sd  = 1
        weight_Mms = 1
        weight_Cms = 1
    elif extension == ".sdrv":
        loader = lpl.speakerSim_lp_loader
        weight_Le  = 1
        weight_Sd  = 1
        weight_Mms = 1
        weight_Cms = 1
    elif extension == ".wdr":
        loader = lpl.winSd_lp_loader
        weight_Le  = 1
        weight_Sd  = 1
        weight_Mms = 1
        weight_Cms = 1
    elif extension == ".bastaelement":
        loader = lpl.basta_lp_loader
        weight_Le  = 1
        weight_Sd  = 1
        weight_Mms = 1
        weight_Cms = 1
    elif extension == ".txt":
        with open(lpmfile, 'r') as file:
            first_line = file.readline().strip()
        if first_line == 'Electrical Parameters':
            loader = lpl.klippel_lp_loader
            weight_Le  = 1e-3
            weight_Sd  = 1e-4
            weight_Mms = 1e-3
            weight_Cms = 1e-3
        else:
            loader = lpl.hornResp_lp_loader
            weight_Le  = 1e-3
            weight_Sd  = 1e-4
            weight_Mms = 1e-3
            weight_Cms = 1
    
    # create driver object
    data = loader(lpmfile)
    Le = data["Le"] * weight_Le
    Re = data["Re"]
    Cms = data["Cms"] * weight_Cms
    Mms = data["Mms"] * weight_Mms
    Rms = data["Rms"]
    Bl = data["Bl"]
    Sd = data["Sd"] * weight_Sd
    
    if LeZero is True:
        Le = 1e-12     # otherwise it doesn't work with circuitSolver()
    
    
    if number_of_drivers > 1:
        if wiring == 'parallel':
            n = number_of_drivers
            drv = electroAcousticDriver(U, Le/n, Re/n, Cms/n, Mms*n, 
                                        Rms*n, Bl, Sd*n, freq_array, c, rho)
        elif wiring == 'series':
            n = number_of_drivers
            drv = electroAcousticDriver(U, Le*n, Re*n, Cms/n, 
                                        Mms*n, Rms*n, Bl*n, Sd*n, 
                                        freq_array, c, rho)
        else:
            ValueError("'wiring' must be either 'parallel' or 'series'.")
    else:
        drv = electroAcousticDriver(U, Le, Re, Cms, Mms, Rms, 
                                    Bl, Sd, freq_array, c, rho)
    return drv
    
    
    
